# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest
from pairtools.lib import pairsam_format

testdir = os.path.dirname(os.path.realpath(__file__))
mock_pairsam_path = os.path.join(testdir, "data", "mock.pairsam")
mock_chromsizes_path = os.path.join(testdir, "data", "mock.chrom.sizes")


def test_preserve():
    try:
        result = subprocess.check_output(
            ["python", "-m", "pairtools", "select", "True", mock_pairsam_path],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]
    assert all(l in pairsam_body for l in output_body)


def test_equal():
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "select",
                '(pair_type == "RU") or (pair_type == "UR") or (pair_type == "UU")',
                mock_pairsam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e
    print(result)

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]

    assert all(l.split("\t")[7] in ["RU", "UR", "UU"] for l in output_body)
    assert all(
        l in output_body for l in pairsam_body if l.split("\t")[7] in ["RU", "UR", "UU"]
    )


def test_csv():
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "select",
                'csv_match(pair_type, "RU,UR,UU")',
                mock_pairsam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e
    print(result)

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]

    assert all(l.split("\t")[7] in ["RU", "UR", "UU"] for l in output_body)
    assert all(
        l in output_body for l in pairsam_body if l.split("\t")[7] in ["RU", "UR", "UU"]
    )


def test_wildcard():
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "select",
                'wildcard_match(pair_type, "*U")',
                mock_pairsam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e
    print(result)

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]

    assert all(l.split("\t")[7] in ["NU", "MU", "RU", "UU"] for l in output_body)
    assert all(
        l in output_body
        for l in pairsam_body
        if l.split("\t")[7] in ["NU", "MU", "RU", "UU"]
    )


def test_regex():
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "select",
                'regex_match(pair_type, "[NM]U")',
                mock_pairsam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e
    print(result)

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]

    assert all(l.split("\t")[7] in ["NU", "MU"] for l in output_body)
    assert all(
        l in output_body for l in pairsam_body if l.split("\t")[7] in ["NU", "MU"]
    )


def test_chrom_subset():
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "select",
                "True",
                "--chrom-subset",
                mock_chromsizes_path,
                mock_pairsam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    pairsam_body = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]
    output_header = [
        l.strip() for l in result.split("\n") if l.startswith("#") and l.strip()
    ]

    chroms_from_chrom_field = [
        l.strip().split()[1:]
        for l in result.split("\n")
        if l.startswith("#chromosomes:")
    ][0]

    assert set(chroms_from_chrom_field) == set(["chr1", "chr2"])

    chroms_from_chrom_sizes = [
        l.strip().split()[1] for l in result.split("\n") if l.startswith("#chromsize:")
    ]

    assert set(chroms_from_chrom_sizes) == set(["chr1", "chr2"])


def test_remove_columns():
    """Test removal of columns from the file
    Example run:
    pairtools select True --remove-columns sam1,sam2 tests/data/mock.pairsam
    """

    mock_pairs_path = os.path.join(testdir, "data", "mock.pairsam")
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "select",
                "True",
                "--remove-columns",
                "sam1,sam2",
                mock_pairs_path,
            ],
        ).decode("ascii")

    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check if the columns are removed properly:
    pairsam_header = [l.strip() for l in result.split("\n") if l.startswith("#")]
    for l in pairsam_header:
        if l.startswith("#columns:"):
            line = l.strip()
            assert (
                line
                == "#columns: readID chrom1 pos1 chrom2 pos2 strand1 strand2 pair_type"
            )

    # check that the pairs got assigned properly
    for l in result.split("\n"):
        if l.startswith("#") or not l:
            continue

        assert len(l.split(pairsam_format.PAIRSAM_SEP)) == 8



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import click

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.dedup import mark_split_pair_as_dup


UTIL_NAME = "pairtools_markasdup"


@cli.command()
@click.argument("pairsam_path", type=str, required=False)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output .pairsam file."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@common_io_options
def markasdup(pairsam_path, output, **kwargs):
    """Tag all pairs in the input file as duplicates.

    Change the type of all pairs inside a .pairs/.pairsam file to DD. If sam
    entries are present, change the pair type in the Yt SAM tag to 'Yt:Z:DD'.

    PAIRSAM_PATH : input .pairs/.pairsam file. If the path ends with .gz, the
    input is gzip-decompressed. By default, the input is read from stdin.
    """
    markasdup_py(pairsam_path, output, **kwargs)


def markasdup_py(pairsam_path, output, **kwargs):
    instream = fileio.auto_open(
        pairsam_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    outstream.writelines((l + "\n" for l in header))

    for line in body_stream:
        cols = line.rstrip('\n').split(pairsam_format.PAIRSAM_SEP)
        mark_split_pair_as_dup(cols)

        outstream.write(pairsam_format.PAIRSAM_SEP.join(cols))
        outstream.write("\n")

    if instream != sys.stdin:
        instream.close()
    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    markasdup()



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import numpy as np
import yaml

testdir = os.path.dirname(os.path.realpath(__file__))


def test_mock_pairsam():
    mock_pairsam_path = os.path.join(testdir, "data", "mock.4stats.pairs")
    try:
        result = subprocess.check_output(
            ["python", "-m", "pairtools", "stats", "--yaml", mock_pairsam_path],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    stats = yaml.safe_load(result)

    # for k in stats["no_filter"]:
    #     try:
    #         stats["no_filter"][k] = int(stats["no_filter"][k])
    #     except (ValueError, TypeError):
    #         stats["no_filter"][k] = float(stats["no_filter"][k])

    assert stats["no_filter"]["total"] == 9
    assert stats["no_filter"]["total_single_sided_mapped"] == 2
    assert stats["no_filter"]["total_mapped"] == 6
    assert stats["no_filter"]["total_dups"] == 1
    assert stats["no_filter"]["cis"] == 3
    assert stats["no_filter"]["trans"] == 2
    assert stats["no_filter"]["pair_types"]["UU"] == 4
    assert stats["no_filter"]["pair_types"]["NU"] == 1
    assert stats["no_filter"]["pair_types"]["WW"] == 1
    assert stats["no_filter"]["pair_types"]["UR"] == 1
    assert stats["no_filter"]["pair_types"]["MU"] == 1
    assert stats["no_filter"]["pair_types"]["DD"] == 1
    assert stats["no_filter"]["chrom_freq"]["chr1/chr2"] == 1
    assert stats["no_filter"]["chrom_freq"]["chr1/chr1"] == 3
    assert stats["no_filter"]["chrom_freq"]["chr2/chr3"] == 1
    for orientation in ("++", "+-", "-+", "--"):
        s = stats["no_filter"]["dist_freq"][orientation]
        for k, val in s.items():
            if orientation == "++" and k in [1, 2, 32]:
                assert s[k] == 1
            else:
                assert s[k] == 0

    assert stats["no_filter"]["summary"]["frac_cis"] == 0.6
    assert stats["no_filter"]["summary"]["frac_cis_1kb+"] == 0
    assert stats["no_filter"]["summary"]["frac_cis_2kb+"] == 0
    assert stats["no_filter"]["summary"]["frac_cis_4kb+"] == 0
    assert stats["no_filter"]["summary"]["frac_cis_10kb+"] == 0
    assert stats["no_filter"]["summary"]["frac_cis_20kb+"] == 0
    assert stats["no_filter"]["summary"]["frac_cis_40kb+"] == 0
    assert np.isclose(stats["no_filter"]["summary"]["frac_dups"], 1 / 6)


def test_merge_stats():
    mock_pairsam_path = os.path.join(testdir, "data", "mock.4stats.pairs")
    try:
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "stats",
                "--with-chromsizes",
                mock_pairsam_path,
                "--output",
                "mock.stats",
            ],
        )

        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "stats",
                "--no-chromsizes",
                mock_pairsam_path,
                "--output",
                "mock.no_chromsizes.stats",
            ],
        )
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "stats",
                "mock.stats",
                "mock.stats",
                "--merge",
                "--output",
                "mock.merged_chromsizes.stats",
            ],
        )
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "stats",
                "mock.stats",
                "mock.no_chromsizes.stats",
                "--merge",
                "--output",
                "mock.merged_mixed.stats",
            ],
        )
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "stats",
                "mock.no_chromsizes.stats",
                "mock.no_chromsizes.stats",
                "--merge",
                "--output",
                "mock.merged_no_chromsizes.stats",
            ],
        )
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest
import pandas as pd
import io

testdir = os.path.dirname(os.path.realpath(__file__))


def test_scaling():
    mock_pairsam_path = os.path.join(testdir, "data", "mock.pairsam")
    try:
        result = subprocess.check_output(
            ["python", "-m", "pairtools", "scaling", mock_pairsam_path],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    output = pd.read_csv(io.StringIO(result), sep="\t", header=0)

    # All the pairs, even "!" are counted as present because we don't provide regions
    assert output["n_pairs"].sum() == 8



#!/usr/bin/env python
# -*- coding: utf-8  -*-
import sys
import ast
import warnings
import pathlib

import click

from ..lib import fileio, pairsam_format, headerops, dedup
from . import cli, common_io_options

from ..lib.filterbycov import streaming_filterbycov
from ..lib.stats import PairCounter


UTIL_NAME = "pairtools_filterbycov"

######################################
## TODO: - output stats after filtering
## edit/update mark as dup to mark as multi
###################################


@cli.command()
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file for pairs from low coverage regions."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@click.option(
    "--output-highcov",
    type=str,
    default="",
    help="output file for pairs from high coverage regions."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " If the path is the same as in --output or -, output duplicates together "
    " with deduped pairs. By default, duplicates are dropped.",
)
@click.option(
    "--output-unmapped",
    type=str,
    default="",
    help="output file for unmapped pairs. "
    "If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed. "
    "If the path is the same as in --output or -, output unmapped pairs together "
    "with deduped pairs. If the path is the same as --output-highcov, "
    "output unmapped reads together. By default, unmapped pairs are dropped.",
)
@click.option(
    "--output-stats",
    type=str,
    default="",
    help="output file for statistics of multiple interactors. "
    " If file exists, it will be open in the append mode."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, statistics are not printed.",
)
@click.option(
    "--max-cov", type=int, default=8, help="The maximum allowed coverage per region."
)
@click.option(
    "--max-dist",
    type=int,
    default=500,
    help="The resolution for calculating coverage. For each pair, the local "
    "coverage around each end is calculated as (1 + the number of neighbouring "
    "pairs within +/- max_dist bp) ",
)
@click.option(
    "--method",
    type=click.Choice(["max", "sum"]),
    default="max",
    help="calculate the number of neighbouring pairs as either the sum or the max"
    " of the number of neighbours on the two sides",
    show_default=True,
)
@click.option(
    "--sep",
    type=str,
    default=pairsam_format.PAIRSAM_SEP_ESCAPE,
    help=r"Separator (\t, \v, etc. characters are " "supported, pass them in quotes) ",
)
@click.option(
    "--comment-char", type=str, default="#", help="The first character of comment lines"
)
@click.option(
    "--send-header-to",
    type=click.Choice(["lowcov", "highcov", "both", "none"]),
    default="both",
    help="Which of the outputs should receive header and comment lines",
)
@click.option(
    "--c1",
    type=int,
    default=pairsam_format.COL_C1,
    help="Chrom 1 column; default {}".format(pairsam_format.COL_C1),
)
@click.option(
    "--c2",
    type=int,
    default=pairsam_format.COL_C2,
    help="Chrom 2 column; default {}".format(pairsam_format.COL_C2),
)
@click.option(
    "--p1",
    type=int,
    default=pairsam_format.COL_P1,
    help="Position 1 column; default {}".format(pairsam_format.COL_P1),
)
@click.option(
    "--p2",
    type=int,
    default=pairsam_format.COL_P2,
    help="Position 2 column; default {}".format(pairsam_format.COL_P2),
)
@click.option(
    "--s1",
    type=int,
    default=pairsam_format.COL_S1,
    help="Strand 1 column; default {}".format(pairsam_format.COL_S1),
)
@click.option(
    "--s2",
    type=int,
    default=pairsam_format.COL_S2,
    help="Strand 2 column; default {}".format(pairsam_format.COL_S2),
)
@click.option(
    "--unmapped-chrom",
    type=str,
    default=pairsam_format.UNMAPPED_CHROM,
    help="Placeholder for a chromosome on an unmapped side; default {}".format(
        pairsam_format.UNMAPPED_CHROM
    ),
)
@click.option(
    "--mark-multi",
    is_flag=True,
    help='If specified, duplicate pairs are marked as FF in "pair_type" and '
    "as a duplicate in the sam entries.",
)
@common_io_options
def filterbycov(
    pairs_path,
    output,
    output_highcov,
    output_unmapped,
    output_stats,
    max_dist,
    max_cov,
    method,
    sep,
    comment_char,
    send_header_to,
    c1,
    c2,
    p1,
    p2,
    s1,
    s2,
    unmapped_chrom,
    mark_multi,
    **kwargs
):
    """Remove pairs from regions of high coverage.

    Find and remove pairs with >(MAX_COV-1) neighbouring pairs
    within a +/- MAX_DIST bp window around either side. Useful for single-cell
    Hi-C experiments, where coverage is naturally limited by the chromosome
    copy number.

    PAIRS_PATH : input triu-flipped sorted .pairs or .pairsam file.  If the
    path ends with .gz/.lz4, the input is decompressed by bgzip/lz4c.
    By default, the input is read from stdin.
    """
    filterbycov_py(
        pairs_path,
        output,
        output_highcov,
        output_unmapped,
        output_stats,
        max_dist,
        max_cov,
        method,
        sep,
        comment_char,
        send_header_to,
        c1,
        c2,
        p1,
        p2,
        s1,
        s2,
        unmapped_chrom,
        mark_multi,
        **kwargs
    )


def filterbycov_py(
    pairs_path,
    output,
    output_highcov,
    output_unmapped,
    output_stats,
    max_dist,
    max_cov,
    method,
    sep,
    comment_char,
    send_header_to,
    c1,
    c2,
    p1,
    p2,
    s1,
    s2,
    unmapped_chrom,
    mark_multi,
    **kwargs
):

    ## Prepare input, output streams based on selected outputs
    ## Default ouput stream is low-frequency interactors
    sep = ast.literal_eval('"""' + sep + '"""')
    send_header_to_lowcov = send_header_to in ["both", "lowcov"]
    send_header_to_highcov = send_header_to in ["both", "highcov"]

    instream = (
        fileio.auto_open(
            pairs_path,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        if pairs_path
        else sys.stdin
    )
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )
    out_stats_stream = (
        fileio.auto_open(
            output_stats,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output_stats
        else None
    )

    # generate empty PairCounter if stats output is requested:
    out_stat = PairCounter() if output_stats else None

    # output the high-frequency interacting pairs
    if not output_highcov:
        outstream_high = None
    elif output_highcov == "-" or (
        pathlib.Path(output_highcov).absolute() == pathlib.Path(output).absolute()
    ):
        outstream_high = outstream
    else:
        outstream_high = fileio.auto_open(
            output_highcov,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    # output unmapped pairs
    if not output_unmapped:
        outstream_unmapped = None
    elif output_unmapped == "-" or (
        pathlib.Path(output_unmapped).absolute() == pathlib.Path(output).absolute()
    ):
        outstream_unmapped = outstream
    elif (
        pathlib.Path(output_unmapped).absolute()
        == pathlib.Path(output_highcov).absolute()
    ):
        outstream_unmapped = outstream_high
    else:
        outstream_unmapped = fileio.auto_open(
            output_unmapped,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    # prepare file headers
    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)

    # header for low-frequency interactors
    if send_header_to_lowcov:
        outstream.writelines((l + "\n" for l in header))

    # header for high-frequency interactors
    if send_header_to_highcov and outstream_high and (outstream_high != outstream):
        outstream_high.writelines((l + "\n" for l in header))

    # header for unmapped pairs
    if (
        outstream_unmapped
        and (outstream_unmapped != outstream)
        and (outstream_unmapped != outstream_high)
    ):
        outstream_unmapped.writelines((l + "\n" for l in header))

    # perform filtering of pairs based on low/high-frequency of interaction
    streaming_filterbycov(
        method,
        max_dist,
        max_cov,
        sep,
        c1,
        c2,
        p1,
        p2,
        s1,
        s2,
        unmapped_chrom,
        body_stream,
        outstream,
        outstream_high,
        outstream_unmapped,
        out_stat,
        mark_multi,
    )

    ## FINISHED!
    # save statistics to a file if it was requested: TO BE TESTED
    if out_stat:
        out_stat.save(out_stats_stream)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()

    if outstream_high and (outstream_high != outstream):
        outstream_high.close()

    if (
        outstream_unmapped
        and (outstream_unmapped != outstream)
        and (outstream_unmapped != outstream_high)
    ):
        outstream_unmapped.close()

    if out_stats_stream:
        out_stats_stream.close()


if __name__ == "__main__":
    filterbycov()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
import pipes
import click

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

UTIL_NAME = "pairtools_split"


@cli.command()
@click.argument("pairsam_path", type=str, required=False)
@click.option(
    "--output-pairs",
    type=str,
    default="",
    help="output pairs file."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " If -, pairs are printed to stdout."
    " If not specified, pairs are dropped.",
)
@click.option(
    "--output-sam",
    type=str,
    default="",
    help="output sam file."
    " If the path ends with .bam, the output is compressed into a bam file."
    " If -, sam entries are printed to stdout."
    " If not specified, sam entries are dropped.",
)
@common_io_options
def split(pairsam_path, output_pairs, output_sam, **kwargs):
    """Split a .pairsam file into .pairs and .sam.

    Restore a .sam file from sam1 and sam2 fields of a .pairsam file. Create
    a .pairs file without sam1/sam2 fields.

    PAIRSAM_PATH : input .pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip or lz4c. By default, the input is read from
    stdin.
    """
    split_py(pairsam_path, output_pairs, output_sam, **kwargs)


def split_py(pairsam_path, output_pairs, output_sam, **kwargs):
    instream = fileio.auto_open(
        pairsam_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )

    # Output streams
    if (not output_pairs) and (not output_sam):
        raise ValueError("At least one output (pairs and/or sam) must be specified!")
    if (output_pairs == "-") and (output_sam == "-"):
        raise ValueError("Only one output (pairs or sam) can be printed in stdout!")

    outstream_pairs = None
    outstream_sam = None

    if output_pairs:
        outstream_pairs = fileio.auto_open(
            output_pairs,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
    if output_sam:
        outstream_sam = fileio.auto_open(
            output_sam,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    columns = headerops.extract_column_names(header)

    has_sams = False
    if columns:
        # trust the column order specified in the header
        if ("sam1" in columns) and ("sam2" in columns):
            sam1col = columns.index("sam1")
            sam2col = columns.index("sam2")
            columns.pop(max(sam1col, sam2col))
            columns.pop(min(sam1col, sam2col))
            header = headerops._update_header_entry(
                header, "columns", " ".join(columns)
            )
            has_sams = True
        elif ("sam1" in columns) != ("sam2" in columns):
            raise ValueError(
                "According to the #columns header field only one sam entry is present"
            )
    else:
        # assume that the file has sam columns and follows the pairsam format
        sam1col = pairsam_format.COL_SAM1
        sam2col = pairsam_format.COL_SAM2
        has_sams = True

    if output_pairs:
        outstream_pairs.writelines((l + "\n" for l in header))
    if output_sam:
        outstream_sam.writelines(
            (l[11:].strip() + "\n" for l in header if l.startswith("#samheader:"))
        )

    # Split
    sam1 = None
    sam2 = None
    for line in body_stream:
        cols = line.rstrip('\n').split(pairsam_format.PAIRSAM_SEP)
        if has_sams:
            if sam1col < sam2col:
                sam2 = cols.pop(sam2col)
                sam1 = cols.pop(sam1col)
            else:
                sam1 = cols.pop(sam1col)
                sam2 = cols.pop(sam2col)

        if output_pairs:
            # hard-coded tab separator to follow the DCIC pairs standard
            outstream_pairs.write("\t".join(cols))
            outstream_pairs.write("\n")

        if output_sam and has_sams:
            for col in (sam1, sam2):
                if col != ".":
                    for sam_entry in col.split(pairsam_format.INTER_SAM_SEP):
                        outstream_sam.write(
                            sam_entry.replace(pairsam_format.SAM_SEP, "\t")
                        )
                        outstream_sam.write("\n")

    if output_pairs and outstream_pairs != sys.stdout:
        outstream_pairs.close()

    if output_sam and outstream_sam != sys.stdout:
        outstream_sam.close()


if __name__ == "__main__":
    split()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import io
import sys
import click
import subprocess
import shutil
import warnings

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

UTIL_NAME = "pairtools_sort"


@cli.command()
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output pairs file."
    " If the path ends with .gz or .lz4, the output is compressed by bgzip "
    "or lz4, correspondingly. By default, the output is printed into stdout.",
)
@click.option(
    "--nproc",
    type=int,
    default=8,
    show_default=True,
    help="Number of processes to split the sorting work between.",
)
@click.option(
    "--tmpdir",
    type=str,
    default="",
    help="Custom temporary folder for sorting intermediates.",
)
@click.option(
    "--memory",
    type=str,
    default="2G",
    show_default=True,
    help="The amount of memory used by default.",
)
@click.option(
    "--compress-program",
    type=str,
    default="auto",
    show_default=True,
    help="A binary to compress temporary sorted chunks. "
    "Must decompress input when the flag -d is provided. "
    "Suggested alternatives: gzip, lzop, lz4c, snzip. "
    'If "auto", then use lz4c if available, and gzip '
    "otherwise.",
)
@common_io_options
def sort(pairs_path, output, nproc, tmpdir, memory, compress_program, **kwargs):
    """Sort a .pairs/.pairsam file.

    Sort pairs in the lexicographic order along chrom1 and chrom2, in the
    numeric order along pos1 and pos2 and in the lexicographic order along
    pair_type.

    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip or lz4c, correspondingly. By default, the
    input is read as text from stdin.
    """
    sort_py(pairs_path, output, nproc, tmpdir, memory, compress_program, **kwargs)


def sort_py(pairs_path, output, nproc, tmpdir, memory, compress_program, **kwargs):

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    header = headerops.mark_header_as_sorted(header)

    outstream.writelines((l + "\n" for l in header))

    outstream.flush()

    if compress_program == "auto":
        if shutil.which("lz4c") is not None:
            compress_program = "lz4c"
        else:
            warnings.warn(
                "lz4c is not found. Using gzip for compression of sorted chunks, "
                "which results in a minor decrease in performance. Please install "
                "lz4c for faster sorting."
            )
            compress_program = "gzip"

    command = r"""
        /bin/bash -c 'export LC_COLLATE=C; export LANG=C; sort 
        -k {0},{0} -k {1},{1} -k {2},{2}n -k {3},{3}n -k {4},{4} 
        --stable
        --field-separator=$'\''{5}'\'' 
        {6}
        {7}
        -S {8}
        {9}
        """.replace(
        "\n", " "
    ).format(
        pairsam_format.COL_C1 + 1,
        pairsam_format.COL_C2 + 1,
        pairsam_format.COL_P1 + 1,
        pairsam_format.COL_P2 + 1,
        pairsam_format.COL_PTYPE + 1,
        pairsam_format.PAIRSAM_SEP_ESCAPE,
        " --parallel={} ".format(nproc) if nproc > 0 else " ",
        " --temporary-directory={} ".format(tmpdir) if tmpdir else " ",
        memory,
        (
            " --compress-program={} ".format(compress_program)
            if compress_program
            else " "
        ),
    )
    command += "'"

    with subprocess.Popen(
        command, stdin=subprocess.PIPE, bufsize=-1, shell=True, stdout=outstream
    ) as process:
        stdin_wrapper = io.TextIOWrapper(process.stdin, "utf-8")
        for line in body_stream:
            stdin_wrapper.write(line)
        stdin_wrapper.flush()
        process.communicate()

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    sort()



import numpy as np
import pandas as pd
from scipy import special
from collections.abc import Mapping
import sys
import yaml
from . import fileio
from .select import evaluate_df

from .._logging import get_logger

logger = get_logger()


class PairCounter(Mapping):
    """
    A Counter for Hi-C pairs that accumulates various statistics.
    PairCounter implements two interfaces to access multi-level statistics:
    1. as a nested dict, e.g. pairCounter['pair_types']['LL']
    2. as a flat dict, with the level keys separated by '/', e.g. pairCounter['pair_types/LL']
    Other features:
    -- PairCounters can be saved into/loaded from a file
    -- multiple PairCounters can be merged via addition.
    """

    _SEP = "\t"
    _KEY_SEP = "/"

    def __init__(
        self,
        min_log10_dist=0,
        max_log10_dist=9,
        log10_dist_bin_step=0.25,
        bytile_dups=False,
        filters=None,
        **kwargs,
    ):
        # Define filters and parameters for filters evaluation:
        if filters is not None:
            self.filters = filters
        else:
            self.filters = {"no_filter": ""}
        self.startup_code = kwargs.get("startup_code", "")
        self.type_cast = kwargs.get("type_cast", ())
        self.engine = kwargs.get("engine", "pandas")

        # Define default filter:
        if "no_filter" not in self.filters:
            self.filters["no_filter"] = ""
        self._stat = {key: {} for key in self.filters}

        # some variables used for initialization:
        # genomic distance bining for the ++/--/-+/+- distribution
        self._dist_bins = np.r_[
            0,
            np.round(
                10
                ** np.arange(
                    min_log10_dist, max_log10_dist + 0.001, log10_dist_bin_step
                )
            ).astype(np.int),
        ]

        # establish structure of an empty _stat:
        for key in self.filters:
            self._stat[key]["filter_expression"] = self.filters[key]
            self._stat[key]["total"] = 0
            self._stat[key]["total_unmapped"] = 0
            self._stat[key]["total_single_sided_mapped"] = 0
            # total_mapped = total_dups + total_nodups
            self._stat[key]["total_mapped"] = 0
            self._stat[key]["total_dups"] = 0
            self._stat[key]["total_nodups"] = 0
            ########################################
            # the rest of stats are based on nodups:
            ########################################
            self._stat[key]["cis"] = 0
            self._stat[key]["trans"] = 0
            self._stat[key]["pair_types"] = {}
            # to be removed:
            self._stat[key]["dedup"] = {}

            self._stat[key]["cis_1kb+"] = 0
            self._stat[key]["cis_2kb+"] = 0
            self._stat[key]["cis_4kb+"] = 0
            self._stat[key]["cis_10kb+"] = 0
            self._stat[key]["cis_20kb+"] = 0
            self._stat[key]["cis_40kb+"] = 0
            self._stat[key]["summary"] = dict(
                [
                    ("frac_cis", 0),
                    ("frac_cis_1kb+", 0),
                    ("frac_cis_2kb+", 0),
                    ("frac_cis_4kb+", 0),
                    ("frac_cis_10kb+", 0),
                    ("frac_cis_20kb+", 0),
                    ("frac_cis_40kb+", 0),
                    ("frac_dups", 0),
                    ("complexity_naive", 0),
                ]
            )

            self._stat[key]["chrom_freq"] = {}

            self._stat[key]["dist_freq"] = {
                "+-": {bin.item(): 0 for bin in self._dist_bins},
                "-+": {bin.item(): 0 for bin in self._dist_bins},
                "--": {bin.item(): 0 for bin in self._dist_bins},
                "++": {bin.item(): 0 for bin in self._dist_bins},
            }

            self._stat[key]["chromsizes"] = {}

            # Summaries are derived from other stats and are recalculated on merge

        self._save_bytile_dups = bytile_dups
        if self._save_bytile_dups:
            self._bytile_dups = pd.DataFrame(
                index=pd.MultiIndex(
                    levels=[[], []], codes=[[], []], names=["tile", "parent_tile"]
                )
            )
        self._summaries_calculated = False

    def __getitem__(self, key, filter="no_filter"):
        if isinstance(key, str):
            # let's strip any unintentional '/'
            # from either side of the key
            key = key.strip("/")
            if self._KEY_SEP in key:
                # multi-key to access nested elements
                k_fields = key.split(self._KEY_SEP)
            else:
                # single-key access flat part of PairCounter
                # or to access highest level of hierarchy
                return self._stat[filter][key]
        else:
            # clearly an error:
            raise ValueError("{} is not a valid key: must be str".format(key))

        # K_FIELDS:
        # process multi-key case:
        # in this case key must be in ['pair_types','chrom_freq','dist_freq','dedup']
        # get the first 'k' and keep the remainders in 'k_fields'
        k = k_fields.pop(0)
        if k in ["pair_types", "dedup"]:
            # assert there is only one element in key_fields left:
            # 'pair_types' and 'dedup' treated the same
            if len(k_fields) == 1:
                return self._stat[filter][k][k_fields[0]]
            else:
                raise ValueError(
                    "{} is not a valid key: {} section implies 1 identifier".format(
                        key, k
                    )
                )

        elif k == "chrom_freq":
            # assert remaining key_fields == [chr1, chr2]:
            if len(k_fields) == 2:
                return self._stat[filter][k][tuple(k_fields)]
            else:
                raise ValueError(
                    "{} is not a valid key: {} section implies 2 identifiers".format(
                        key, k
                    )
                )

        elif k == "dist_freq":
            # assert that last element of key_fields is the 'directions'
            # THIS IS DONE FOR CONSISTENCY WITH .stats FILE
            # SHOULD THAT BE CHANGED IN .stats AND HERE AS WELL?
            if len(k_fields) == 2:
                # assert 'dirs' in ['++','--','+-','-+']
                dirs = k_fields.pop()
                # there is only genomic distance range of the bin that's left:
                (bin_range,) = k_fields
                # extract left border of the bin "1000000+" or "1500-6000":
                dist_bin_left = (
                    bin_range.strip("+")
                    if bin_range.endswith("+")
                    else bin_range.split("-")[0]
                )
                # get the index of that bin:
                bin_idx = (
                    np.searchsorted(self._dist_bins, int(dist_bin_left), "right") - 1
                )
                # store corresponding value:
                return self._stat[filter]["dist_freq"][dirs][bin_idx]
            else:
                raise ValueError(
                    "{} is not a valid key: {} section implies 2 identifiers".format(
                        key, k
                    )
                )
        else:
            raise ValueError("{} is not a valid key".format(k))

    def __iter__(self):
        return iter(self._stat)

    def __len__(self):
        return len(self._stat)

    def calculate_summaries(self):
        """calculate summary statistics (fraction of cis pairs at different cutoffs,
        complexity estimate) based on accumulated counts. Results are saved into
        self._stat["filter_name"]['summary"]
        """
        for key in self.filters.keys():
            self._stat[key]["summary"]["frac_dups"] = (
                (self._stat[key]["total_dups"] / self._stat[key]["total_mapped"])
                if self._stat[key]["total_mapped"] > 0
                else 0
            )

            for cis_count in (
                "cis",
                "cis_1kb+",
                "cis_2kb+",
                "cis_4kb+",
                "cis_10kb+",
                "cis_20kb+",
                "cis_40kb+",
            ):
                self._stat[key]["summary"][f"frac_{cis_count}"] = (
                    (self._stat[key][cis_count] / self._stat[key]["total_nodups"])
                    if self._stat[key]["total_nodups"] > 0
                    else 0
                )

            self._stat[key]["summary"][
                "complexity_naive"
            ] = estimate_library_complexity(
                self._stat[key]["total_mapped"], self._stat[key]["total_dups"], 0
            )
            if key == "no_filter" and self._save_bytile_dups:
                # Estimate library complexity with information by tile, if provided:
                if self._bytile_dups.shape[0] > 0:
                    self._stat[key]["dups_by_tile_median"] = int(
                        round(
                            self._bytile_dups["dup_count"].median()
                            * self._bytile_dups.shape[0]
                        )
                    )
                if "dups_by_tile_median" in self._stat[key].keys():
                    self._stat[key]["summary"][
                        "complexity_dups_by_tile_median"
                    ] = estimate_library_complexity(
                        self._stat[key]["total_mapped"],
                        self._stat[key]["total_dups"],
                        self._stat[key]["total_dups"]
                        - self._stat[key]["dups_by_tile_median"],
                    )

            self._summaries_calculated = True

    @classmethod
    def from_file(cls, file_handle):
        """create instance of PairCounter from file
        Parameters
        ----------
        file_handle: file handle
        Returns
        -------
        PairCounter
            new PairCounter filled with the contents of the input file
        """
        # fill in from file - file_handle:
        default_filter = "no_filter"
        stat_from_file = cls()
        for l in file_handle:
            fields = l.strip().split(cls._SEP)
            if len(fields) == 0:
                # skip empty lines:
                continue
            if len(fields) != 2:
                # expect two _SEP separated values per line:
                raise fileio.ParseError(
                    "{} is not a valid stats file".format(file_handle.name)
                )
            # extract key and value, then split the key:
            putative_key, putative_val = fields[0], fields[1]
            key_fields = putative_key.split(cls._KEY_SEP)
            # we should impose a rigid structure of .stats or redo it:
            if len(key_fields) == 1:
                key = key_fields[0]
                if key in stat_from_file._stat[default_filter]:
                    stat_from_file._stat[default_filter][key] = int(fields[1])
                else:
                    raise fileio.ParseError(
                        "{} is not a valid stats file: unknown field {} detected".format(
                            file_handle.name, key
                        )
                    )
            else:
                # in this case key must be in ['pair_types','chrom_freq','dist_freq','dedup', 'summary']
                # get the first 'key' and keep the remainders in 'key_fields'
                key = key_fields.pop(0)
                if key in ["pair_types", "dedup", "summary", "chromsizes"]:
                    # assert there is only one element in key_fields left:
                    # 'pair_types', 'dedup', 'summary' and 'chromsizes' treated the same
                    if len(key_fields) == 1:
                        try:
                            stat_from_file._stat[default_filter][key][
                                key_fields[0]
                            ] = int(fields[1])
                        except ValueError:
                            stat_from_file._stat[default_filter][key][
                                key_fields[0]
                            ] = float(fields[1])
                    else:
                        raise fileio.ParseError(
                            "{} is not a valid stats file: {} section implies 1 identifier".format(
                                file_handle.name, key
                            )
                        )

                elif key == "chrom_freq":
                    # assert remaining key_fields == [chr1, chr2]:
                    if len(key_fields) == 2:
                        stat_from_file._stat[default_filter][key][
                            tuple(key_fields)
                        ] = int(fields[1])
                    else:
                        raise fileio.ParseError(
                            "{} is not a valid stats file: {} section implies 2 identifiers".format(
                                file_handle.name, key
                            )
                        )

                elif key == "dist_freq":
                    # assert that last element of key_fields is the 'directions'
                    if len(key_fields) == 2:
                        # assert 'dirs' in ['++','--','+-','-+']
                        dirs = key_fields.pop()
                        # there is only genomic distance range of the bin that's left:
                        (bin_range,) = key_fields
                        # extract left border of the bin "1000000+" or "1500-6000":
                        dist_bin_left = (
                            bin_range.strip("+")
                            if bin_range.endswith("+")
                            else bin_range.split("-")[0]
                        )
                        # get the index of that bin:
                        bin_idx = (
                            np.searchsorted(
                                stat_from_file._dist_bins, int(dist_bin_left), "right"
                            )
                            - 1
                        )
                        # store corresponding value:
                        stat_from_file._stat[default_filter][key][dirs][bin_idx] = int(
                            fields[1]
                        )
                    else:
                        raise fileio.ParseError(
                            "{} is not a valid stats file: {} section implies 2 identifiers".format(
                                file_handle.name, key
                            )
                        )
                else:
                    raise fileio.ParseError(
                        "{} is not a valid stats file: unknown field {} detected".format(
                            file_handle.name, key
                        )
                    )
        # return PairCounter from a non-empty dict:
        return stat_from_file

    @classmethod
    def from_yaml(cls, file_handle):
        """create instance of PairCounter from file
        Parameters
        ----------
        file_handle: file handle
        Returns
        -------
        PairCounter
            new PairCounter filled with the contents of the input file
        """
        # fill in from file - file_handle:
        stat_from_file = cls()

        stat = yaml.safe_load(file_handle)
        for key, filter in stat.items():
            chromdict = {}
            for chroms in stat[key]["chrom_freq"].keys():
                chromdict[tuple(chroms.split(cls._KEY_SEP))] = stat[key]["chrom_freq"][
                    chroms
                ]
            stat[key]["chrom_freq"] = chromdict
        stat_from_file._stat = stat
        return stat_from_file

    def add_pair(
        self,
        chrom1,
        pos1,
        strand1,
        chrom2,
        pos2,
        strand2,
        pair_type,
        filter="no_filter",
    ):
        """Gather statistics for a Hi-C pair and add to the PairCounter.
        Parameters
        ----------
        chrom1: str
            chromosome of the first read
        pos1: int
            position of the first read
        strand1: str
            strand of the first read
        chrom2: str
            chromosome of the first read
        pos2: int
            position of the first read
        strand2: str
            strand of the first read
        pair_type: str
            type of the mapped pair of reads
        """

        self._stat[filter]["total"] += 1
        # collect pair type stats including DD:
        self._stat[filter]["pair_types"][pair_type] = (
            self._stat[filter]["pair_types"].get(pair_type, 0) + 1
        )
        if chrom1 == "!" and chrom2 == "!":
            self._stat[filter]["total_unmapped"] += 1
        elif chrom1 != "!" and chrom2 != "!":
            self._stat[filter]["total_mapped"] += 1
            # only mapped ones can be duplicates:
            if pair_type == "DD":
                self._stat[filter]["total_dups"] += 1
            else:
                self._stat[filter]["total_nodups"] += 1
                self._stat[filter]["chrom_freq"][(chrom1, chrom2)] = (
                    self._stat[filter]["chrom_freq"].get((chrom1, chrom2), 0) + 1
                )

                if chrom1 == chrom2:
                    self._stat[filter]["cis"] += 1
                    dist = np.abs(pos2 - pos1)
                    bin = self._dist_bins[
                        np.searchsorted(self._dist_bins, dist, "right") - 1
                    ]
                    self._stat[filter]["dist_freq"][strand1 + strand2][bin] += 1
                    if dist >= 1000:
                        self._stat[filter]["cis_1kb+"] += 1
                    if dist >= 2000:
                        self._stat[filter]["cis_2kb+"] += 1
                    if dist >= 4000:
                        self._stat[filter]["cis_4kb+"] += 1
                    if dist >= 10000:
                        self._stat[filter]["cis_10kb+"] += 1
                    if dist >= 20000:
                        self._stat[filter]["cis_20kb+"] += 1
                    if dist >= 40000:
                        self._stat[filter]["cis_40kb+"] += 1

                else:
                    self._stat[filter]["trans"] += 1
        else:
            self._stat[filter]["total_single_sided_mapped"] += 1

    def add_pairs_from_dataframe(self, df, unmapped_chrom="!"):
        """Gather statistics for Hi-C pairs in a dataframe and add to the PairCounter.

        Parameters
        ----------
        df: pd.DataFrame
            DataFrame with pairs. Needs to have columns:
                'chrom1', 'pos1', 'chrom2', 'pos2', 'strand1', 'strand2', 'pair_type'
        """
        for key in self.filters.keys():
            if key == "no_filter":
                df_filtered = df.copy()
            else:
                condition = self.filters[key]
                filter_passed = evaluate_df(
                    df,
                    condition,
                    type_cast=self.type_cast,
                    startup_code=self.startup_code,
                    engine=self.engine,
                )
                df_filtered = df.loc[filter_passed, :].reset_index(drop=True)
            total_count = df_filtered.shape[0]
            self._stat[key]["total"] += total_count

            # collect pair type stats including DD:
            for pair_type, type_count in (
                df_filtered["pair_type"].value_counts().items()
            ):
                self._stat[key]["pair_types"][pair_type] = (
                    self._stat[key]["pair_types"].get(pair_type, 0) + type_count
                )

            # Count the unmapped by the "unmapped" chromosomes (debatable, as WW are also marked as ! and they might be mapped):
            unmapped_count = np.logical_and(
                df_filtered["chrom1"] == unmapped_chrom,
                df_filtered["chrom2"] == unmapped_chrom,
            ).sum()
            self._stat[key]["total_unmapped"] += int(unmapped_count)

            # Count the mapped:
            df_mapped = df_filtered.loc[
                (df_filtered["chrom1"] != unmapped_chrom)
                & (df_filtered["chrom2"] != unmapped_chrom),
                :,
            ]
            mapped_count = df_mapped.shape[0]

            self._stat[key]["total_mapped"] += mapped_count
            self._stat[key]["total_single_sided_mapped"] += int(
                total_count - (mapped_count + unmapped_count)
            )

            # Count the duplicates:
            if "duplicate" in df_mapped.columns:
                mask_dups = df_mapped["duplicate"]
            else:
                mask_dups = df_mapped["pair_type"] == "DD"
            df_dups = df_mapped[mask_dups]
            dups_count = df_dups.shape[0]
            self._stat[key]["total_dups"] += int(dups_count)
            self._stat[key]["total_nodups"] += int(mapped_count - dups_count)

            df_nodups = df_mapped.loc[~mask_dups, :]
            mask_cis = df_nodups["chrom1"] == df_nodups["chrom2"]
            df_cis = df_nodups.loc[mask_cis, :].copy()

            # Count pairs per chromosome:
            for (chrom1, chrom2), chrom_count in (
                df_nodups[["chrom1", "chrom2"]].value_counts().items()
            ):
                self._stat[key]["chrom_freq"][(chrom1, chrom2)] = (
                    self._stat[key]["chrom_freq"].get((chrom1, chrom2), 0) + chrom_count
                )

            # Count cis-trans by pairs:

            self._stat[key]["cis"] += df_cis.shape[0]
            self._stat[key]["trans"] += df_nodups.shape[0] - df_cis.shape[0]
            dist = np.abs(df_cis["pos2"].values - df_cis["pos1"].values)

            df_cis.loc[:, "bin_idx"] = (
                np.searchsorted(self._dist_bins, dist, "right") - 1
            )
            for (strand1, strand2, bin_id), strand_bin_count in (
                df_cis[["strand1", "strand2", "bin_idx"]].value_counts().items()
            ):
                self._stat[key]["dist_freq"][strand1 + strand2][
                    self._dist_bins[bin_id].item()
                ] += strand_bin_count
            self._stat[key]["cis_1kb+"] += int(np.sum(dist >= 1000))
            self._stat[key]["cis_2kb+"] += int(np.sum(dist >= 2000))
            self._stat[key]["cis_4kb+"] += int(np.sum(dist >= 4000))
            self._stat[key]["cis_10kb+"] += int(np.sum(dist >= 10000))
            self._stat[key]["cis_20kb+"] += int(np.sum(dist >= 20000))
            self._stat[key]["cis_40kb+"] += int(np.sum(dist >= 40000))

            ### Add by-tile dups
            if key == "no_filter" and self._save_bytile_dups and (df_dups.shape[0] > 0):
                bytile_dups = analyse_bytile_duplicate_stats(df_dups)
                self._bytile_dups = self._bytile_dups.add(
                    bytile_dups, fill_value=0
                ).astype(int)

    def add_chromsizes(self, chromsizes):
        """Add chromsizes field to the output stats
        Parameters
        ----------
        chromsizes: Dataframe with chromsizes, read by headerops.chromsizes
        """

        chromsizes = chromsizes.to_dict()
        for filter in self._stat.keys():
            self._stat[filter]["chromsizes"] = chromsizes
        return

    def __add__(self, other, filter="no_filter"):
        # both PairCounter are implied to have a list of common fields:
        #
        # 'total', 'total_unmapped', 'total_single_sided_mapped', 'total_mapped',
        # 'cis', 'trans', 'pair_types', 'cis_1kb+', 'cis_2kb+',
        # 'cis_10kb+', 'cis_20kb+', 'chrom_freq', 'dist_freq', 'dedup'
        #
        # If 'chromsizes' are present, they must be identical
        #
        # initialize empty PairCounter for the result of summation:
        sum_stat = PairCounter()
        # use the empty PairCounter to iterate over:
        for k, v in sum_stat._stat[filter].items():
            if k != "chromsizes" and (
                k not in self._stat[filter] or k not in other._stat[filter]
            ):
                # Skip any missing fields and warn
                logger.warning(
                    f"{k} not found in at least one of the input stats, skipping"
                )
                continue
            # not nested fields are summed trivially:
            if isinstance(v, int):
                sum_stat._stat[filter][k] = (
                    self._stat[filter][k] + other._stat[filter][k]
                )
            # sum nested dicts/arrays in a context dependet manner:
            else:
                if k in ["pair_types", "dedup", "summary"]:
                    # handy function for summation of a pair of dicts:
                    # https://stackoverflow.com/questions/10461531/merge-and-sum-of-two-dictionaries
                    sum_dicts = lambda dict_x, dict_y: {
                        key: dict_x.get(key, 0) + dict_y.get(key, 0)
                        for key in set(dict_x) | set(dict_y)
                    }
                    # sum a pair of corresponding dicts:
                    sum_stat._stat[filter][k] = sum_dicts(
                        self._stat[filter][k], other._stat[filter][k]
                    )
                elif k == "chrom_freq":
                    # union list of keys (chr1,chr2) with potential duplicates:
                    union_keys_with_dups = list(self._stat[filter][k].keys()) + list(
                        other._stat[filter][k].keys()
                    )
                    # dict.fromkeys will take care of keys' order and duplicates in a consistent manner:
                    # https://stackoverflow.com/questions/1720421/how-to-concatenate-two-lists-in-python
                    # last comment to the 3rd Answer
                    sum_stat._stat[filter][k] = dict.fromkeys(union_keys_with_dups)
                    # perform a summation:
                    for union_key in sum_stat._stat[filter][k]:
                        sum_stat._stat[filter][k][union_key] = self._stat[filter][
                            k
                        ].get(union_key, 0) + other._stat[filter][k].get(union_key, 0)
                elif k == "dist_freq":
                    for dirs in sum_stat[k]:

                        from functools import reduce

                        def reducer(accumulator, element):
                            for key, value in element.items():
                                accumulator[key] = accumulator.get(key, 0) + value
                            return accumulator

                        sum_stat[k][dirs] = reduce(
                            reducer,
                            [self._stat[filter][k][dirs], other._stat[filter][k][dirs]],
                            {},
                        )
                        # sum_stat[k][dirs] = self._stat[filter][k][dirs] + other._stat[filter][k][dirs]
                elif k == "chromsizes":
                    if k in self._stat[filter] and k in other._stat[filter]:
                        if self._stat[filter][k] == other._stat[filter][k]:
                            sum_stat._stat[filter][k] = self._stat[filter][k]
                        elif (
                            len(self._stat[filter][k]) == 0
                            or len(other._stat[filter][k]) == 0
                        ):
                            logger.warning(
                                "One of the stats has no chromsizes recorded,"
                                "writing the one that is present to the output"
                            )
                            if len(self._stat[filter][k]) > 0:
                                sum_stat._stat[filter][k] = self._stat[filter][k]
                            else:
                                sum_stat._stat[filter][k] = other._stat[filter][k]
                        else:
                            raise ValueError(
                                "Can't merge stats with different chromsizes"
                            )
                    else:
                        logger.warning(
                            "One or both stats don't have chromsizes recorded"
                        )

        return sum_stat

    # we need this to be able to sum(list_of_PairCounters)
    def __radd__(self, other):
        if other == 0:
            return self
        else:
            return self.__add__(other)

    def flatten(self, filter="no_filter"):
        """return a flattened dict (formatted same way as .stats file)
        Performed for a single filter."""
        # dict for flat store:
        flat_stat = {}

        # Storing statistics
        for k, v in self._stat[filter].items():
            if isinstance(v, int):
                flat_stat[k] = v
            # store nested dicts/arrays in a context dependet manner:
            # nested categories are stored only if they are non-trivial
            else:
                if (k == "dist_freq") and v:
                    for i in range(len(self._dist_bins)):
                        for dirs, freqs in v.items():
                            # last bin is treated differently: "100000+" vs "1200-3000":
                            if i != len(self._dist_bins) - 1:
                                dist = self._dist_bins[i]
                                dist_next = self._dist_bins[i + 1]
                                formatted_key = self._KEY_SEP.join(
                                    ["{}", "{}-{}", "{}"]
                                ).format(k, dist, dist_next, dirs)
                            else:
                                formatted_key = self._KEY_SEP.join(
                                    ["{}", "{}+", "{}"]
                                ).format(k, dist, dirs)
                            # store key,value pair:
                            flat_stat[formatted_key] = freqs[dist]
                elif (k in ["pair_types", "dedup", "chromsizes"]) and v:
                    # 'pair_types' and 'dedup' are simple dicts inside,
                    # treat them the exact same way:
                    for k_item, freq in v.items():
                        formatted_key = self._KEY_SEP.join(["{}", "{}"]).format(
                            k, k_item
                        )
                        # store key,value pair:
                        flat_stat[formatted_key] = freq
                elif (k == "chrom_freq") and v:
                    for (chrom1, chrom2), freq in v.items():
                        formatted_key = self._KEY_SEP.join(["{}", "{}", "{}"]).format(
                            k, chrom1, chrom2
                        )
                        # store key,value pair:
                        flat_stat[formatted_key] = freq
                elif (k == "summary") and v:
                    for key, frac in v.items():
                        formatted_key = self._KEY_SEP.join(["{}", "{}"]).format(k, key)
                        # store key,value pair:
                        flat_stat[formatted_key] = frac

        # return flattened dict
        return flat_stat

    def format_yaml(self, filter="no_filter"):
        """return a formatted dict (for the yaml output)
        Performed for all filters at once."""

        from copy import deepcopy

        formatted_stat = {key: {} for key in self.filters.keys()}

        # Storing statistics for each filter
        for key in self.filters.keys():
            for k, v in self._stat[key].items():
                if isinstance(v, int):
                    formatted_stat[key][k] = v
                # store nested dicts/arrays in a context dependet manner:
                # nested categories are stored only if they are non-trivial
                else:
                    if (k != "chrom_freq") and v:
                        # simple dicts inside
                        # treat them the exact same way:
                        formatted_stat[key][k] = deepcopy(v)
                    elif (k == "chrom_freq") and v:
                        # need to convert tuples of chromosome names to str
                        freqs = {}
                        for (chrom1, chrom2), freq in sorted(v.items()):
                            freqs[
                                self._KEY_SEP.join(["{}", "{}"]).format(chrom1, chrom2)
                            ] = freq
                            # store key,value pair:
                        formatted_stat[key][k] = deepcopy(freqs)
            # return formatted dict
        return formatted_stat

    def save(self, outstream, yaml=False, filter="no_filter"):
        """save PairCounter to tab-delimited text file.
        Flattened version of PairCounter is stored in the file.
        Parameters
        ----------
        outstream: file handle
        yaml: is output in yaml format or table
        filter: filter to output in tsv mode
        Note
        ----
        The order of the keys is not guaranteed
        Merging several .stats is not associative with respect to key order:
        merge(A,merge(B,C)) != merge(merge(A,B),C).
        Theys shou5ld match exactly, however, when soprted:
        sort(merge(A,merge(B,C))) == sort(merge(merge(A,B),C))
        """

        if not self._summaries_calculated:
            self.calculate_summaries()

        # write flattened version of the PairCounter to outstream,
        # will output all the filters
        if yaml:
            import yaml

            data = self.format_yaml()
            yaml.dump(data, outstream, default_flow_style=False, sort_keys=False)
        else:  # will output a single filter
            data = self.flatten(filter=filter)
            for k, v in data.items():
                outstream.write("{}{}{}\n".format(k, self._SEP, v))

    def save_bytile_dups(self, outstream):
        """save bytile duplication counts to a tab-delimited text file.
        Parameters
        ----------
        outstream: file handle
        """
        if self._save_bytile_dups:
            self._bytile_dups.reset_index().to_csv(outstream, sep="\t", index=False)
        else:
            logger.error("Bytile dups are not calculated, cannot save.")

    def __repr__(self):
        return str(self._stat)


##################
# Other functions:


def do_merge(output, files_to_merge, **kwargs):
    # Parse all stats files.
    stats = []
    for stat_file in files_to_merge:
        f = fileio.auto_open(
            stat_file,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        # use a factory method to instanciate PairCounter
        if kwargs.get("yaml", False):
            stat = PairCounter.from_yaml(f)
        else:
            stat = PairCounter.from_file(f)
        stats.append(stat)
        f.close()

    # combine stats from several files (files_to_merge):
    out_stat = sum(stats)

    # Save merged stats.
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    # save statistics to file ...
    out_stat.save(outstream)

    if outstream != sys.stdout:
        outstream.close()


def estimate_library_complexity(nseq, ndup, nopticaldup=0):
    """Estimate library complexity accounting for optical/clustering duplicates
    Parameters
    ----------
    nseq : int
        Total number of sequences
    ndup : int
        Total number of duplicates
    nopticaldup : int, optional
        Number of non-PCR duplicates, by default 0
    Returns
    -------
    float
        Estimated complexity
    """
    nseq = nseq - nopticaldup
    if nseq == 0:
        logger.warning("Empty of fully duplicated library, can't estimate complexity")
        return 0
    ndup = ndup - nopticaldup
    u = (nseq - ndup) / nseq
    if u == 0:
        logger.warning(
            "All the sequences are duplicates. Do you run complexity estimation on duplicates file?"
        )
        return 0
    seq_to_complexity = special.lambertw(-np.exp(-1 / u) / u).real + 1 / u
    complexity = float(nseq / seq_to_complexity)  # clean np.int64 data type
    return complexity


def analyse_bytile_duplicate_stats(df_dups, tile_dup_regex=False):
    """Count by-tile duplicates
    Parameters
    ----------
    dups : pd.DataFrame
        Dataframe with duplicates that contains pared read IDs
    tile_dup_regex : bool, optional
        See extract_tile_info for details, by default False
    Returns
    -------
    pd.DataFrame
        Grouped multi-indexed dataframe of pairwise by-tile duplication counts
    """

    df_dups = df_dups.copy()

    df_dups["tile"] = extract_tile_info(df_dups["readID"], regex=tile_dup_regex)
    df_dups["parent_tile"] = extract_tile_info(
        df_dups["parent_readID"], regex=tile_dup_regex
    )

    df_dups["same_tile"] = df_dups["tile"] == df_dups["parent_tile"]
    bytile_dups = (
        df_dups.groupby(["tile", "parent_tile"])
        .size()
        .reset_index(name="dup_count")
        .sort_values(["tile", "parent_tile"])
    )
    bytile_dups[["tile", "parent_tile"]] = np.sort(
        bytile_dups[["tile", "parent_tile"]].values, axis=1
    )
    bytile_dups = bytile_dups.groupby(["tile", "parent_tile"]).sum()
    return bytile_dups


def extract_tile_info(series, regex=False):
    """Extract the name of the tile for each read name in the series
    Parameters
    ----------
    series : pd.Series
        Series containing read IDs
    regex : bool, optional
        Regex to extract fields from the read IDs that correspond to tile IDs.
        By default False, uses a faster predefined approach for typical Illumina
        read names
        Example: r"(?:\w+):(?:\w+):(\w+):(\w+):(\w+):(?:\w+):(?:\w+)"
    Returns
    -------
    Series
        Series containing tile IDs as strings
    """
    if regex:
        split = series.str.extractall(regex).unstack().droplevel(1, axis=1)
        if split.shape[1] < 4:
            raise ValueError(
                f"Unable to convert tile names, does your readID have the tile information?\nHint: SRA removes tile information from readID.\nSample of your readIDs:\n{series.head()}"
            )
        return split[0] + ":" + split[1] + ":" + split[2]
    else:
        try:
            split = [":".join(name.split(":")[2:5]) for name in series]
        except:
            raise ValueError(
                f"Unable to convert tile names, does your readID have the tile information?\nHint: SRA removes tile information from readID.\nSample of your readIDs:\n{series.head()}"
            )
        return split


def yaml2pandas(yaml_path):
    """Generate a pandas DataFrame with stats from a yaml file

    Formats the keys within each filter using the PairCounter.flatten() method, to
    achieve same naming as in non-yaml stats files.

    Parameters
    ----------
    yaml_path : str
        Path to a yaml-formatted file with stats

    Returns
    -------
    pd.DataFrame
        Dataframe with filter names in the index and stats in columns
    """
    counter = PairCounter.from_yaml(open(yaml_path, "r"))
    stats = pd.concat(
        [
            pd.DataFrame(counter.flatten(filter=filter), index=[filter])
            for filter in counter.filters
        ]
    )
    return stats



# -*- coding: utf-8 -*-
import os
import sys
import pytest

import subprocess

testdir = os.path.dirname(os.path.realpath(__file__))


def test_generate():
    """Test generation of the header.
    Example run:
    pairtools header generate tests/data/mock.pairsam \
    --chroms-path tests/data/mock.chrom.sizes --pairsam \
    --sam-path tests/data/mock.sam
    """

    mock_sam_path = os.path.join(testdir, "data", "mock.sam")
    mock_pairs_path = os.path.join(testdir, "data", "mock.pairsam")
    mock_chroms_path = os.path.join(testdir, "data", "mock.chrom.sizes")
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "header",
                "generate",
                "--chroms-path",
                mock_chroms_path,
                "--sam-path",
                mock_sam_path,
                "--pairsam",
                mock_pairs_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check if the header got transferred correctly
    sam_header = [l.strip() for l in open(mock_sam_path, "r") if l.startswith("@")]
    pairsam_header = [l.strip() for l in result.split("\n") if l.startswith("#")]
    for l in sam_header:
        assert any([l in l2 for l2 in pairsam_header])



from collections import defaultdict
import sys
import copy
import itertools
import warnings

import numpy as np
import pandas as pd

from .. import __version__
from . import pairsam_format
from .fileio import ParseError

from .._logging import get_logger

logger = get_logger()

PAIRS_FORMAT_VERSION = "1.0.0"
SEP_COLS = " "
SEP_CHROMS = " "
COMMENT_CHAR = "#"


def get_stream_handlers(instream):
    # get peekable buffer for the instream
    readline_f, peek_f = None, None
    if hasattr(instream, "buffer"):
        peek_f = instream.buffer.peek
        readline_f = instream.buffer.readline
    elif hasattr(instream, "peek"):
        peek_f = instream.peek
        readline_f = instream.readline
    else:
        raise ValueError("Cannot find the peek() function of the provided stream!")
    return readline_f, peek_f


def get_header(instream, comment_char=COMMENT_CHAR, ignore_warning=False):
    """Returns a header from the stream and an the reaminder of the stream
    with the actual data.
    Parameters
    ----------
    instream : a file object
        An input stream.
    comment_char : str
        The character prepended to header lines (use '@' when parsing sams,
        '#' when parsing pairsams).
    ignore_warning : bool
        If True, then no warning will be generated if header of pairs file is empty.
    Returns
    -------
    header : list
        The header lines, stripped of terminal spaces and newline characters.
    remainder_stream : stream/file-like object
        Stream with the remaining lines.

    """
    header = []
    if not comment_char:
        raise ValueError("Please, provide a comment char!")
    comment_byte = comment_char.encode()

    readline_f, peek_f = get_stream_handlers(instream)
    current_peek = peek_f(1)
    while current_peek.startswith(comment_byte):
        # consuming a line from buffer guarantees
        # that the remainder of the buffer starts
        # with the beginning of the line.
        line = readline_f()
        if isinstance(line, bytes):
            line = line.decode()
        # append line to header, since it does start with header
        header.append(line.rstrip('\n'))
        # peek into the remainder of the instream
        current_peek = peek_f(1)
    # apparently, next line does not start with the comment
    # return header and the instream, advanced to the beginning of the data

    if len(header) == 0 and not ignore_warning:
        logger.warning(
            "Headerless input, please, add the header by `pairtools header generate` or `pairtools header transfer`"
        )

    return header, instream


def extract_fields(header, field_name, save_rest=False):
    """
    Extract the specified fields from the pairs header and return
    a list of corresponding values, even if a single field was found.
    Additionally, can return the list of intact non-matching entries.
    """

    fields = []
    rest = []
    for l in header:
        if l.lstrip(COMMENT_CHAR).startswith(field_name + ":"):
            fields.append(l.split(":", 1)[1].rstrip('\n').lstrip())
        elif save_rest:
            rest.append(l)

    if save_rest:
        return fields, rest
    else:
        return fields


def extract_column_names(header):
    """
    Extract column names from header lines.
    """
    columns = extract_fields(header, "columns")

    if len(columns) != 0:
        return columns[0].split(SEP_COLS)
    else:
        return []


def validate_cols(stream, columns):
    """
    Validate that the number of columns coincides between stream and columns.
    Checks only the first line in the pairs stream!

    Note that it irreversibly removes the header from the stream.

    Parameters
    ----------
    stream: input stream, body or full .pairs file
    columns: columns to validate against

    Returns
    -------
    True if the number of columns is identical between file and columns
    """

    comment_byte = COMMENT_CHAR.encode()
    readline_f, peek_f = get_stream_handlers(stream)

    current_peek = peek_f(1)
    while current_peek.startswith(comment_byte):
        # consuming a line from buffer guarantees
        # that the remainder of the buffer starts
        # with the beginning of the line.
        line = readline_f()
        # peek into the remainder of the instream
        current_peek = peek_f(1)

    line = readline_f()
    if isinstance(line, bytes):
        line = line.decode()

    ncols_body = len(line.split(pairsam_format.PAIRSAM_SEP))
    ncols_reference = (
        len(columns) if isinstance(columns, list) else columns.split(SEP_COLS)
    )

    return ncols_body == ncols_reference


def validate_header_cols(stream, header):
    """Validate that the number of columns corresponds between the stream and header"""

    columns = extract_column_names(header)
    return validate_cols(stream, header)


def is_empty_header(header):
    if len(header) == 0:
        return True
    if not header[0].startswith("##"):
        return True
    else:
        return False


def extract_chromsizes(header):
    """
    Extract chromosome sizes from header lines.
    """

    chromsizes_str = extract_fields(header, "chromsize")
    chromsizes_str = list(zip(*[s.split(SEP_CHROMS) for s in chromsizes_str]))
    chromsizes = pd.Series(data=chromsizes_str[1], index=chromsizes_str[0]).astype(
        np.int64
    )

    return chromsizes


def get_chromsizes_from_pysam_header(samheader):
    """Convert pysam header to pairtools chromosomes dict (ordered by Python default since 3.7).

    Example of pysam header converted to dict:
    dict([
        ('SQ', [{'SN': 'chr1', 'LN': 248956422},
         {'SN': 'chr10', 'LN': 133797422},
         {'SN': 'chr11', 'LN': 135086622},
         {'SN': 'chr12', 'LN': 133275309}]),
        ('PG', [{'ID': 'bwa', 'PN': 'bwa', 'VN': '0.7.17-r1188', 'CL': 'bwa mem -t 8 -SP -v1 hg38.fa test_1.1.fastq.gz test_2.1.fastq.gz'}])
    ])
    """
    SQs = samheader.to_dict()["SQ"]
    chromsizes = [(sq["SN"], int(sq["LN"])) for sq in SQs]
    return dict(chromsizes)


def get_chromsizes_from_file(chroms_file):
    """
    Produce an "enumeration" of chromosomes based on the list
    of chromosomes
    """
    chrom_sizes = dict()
    with open(chroms_file, "rt") as f:
        for line in f:
            chrom, size = line.strip().split("\t")
            chrom_sizes[chrom] = int(size)

    return chrom_sizes


def get_chromsizes_from_pysam_header(samheader):
    """Convert pysam header to pairtools chromosomes (ordered dict).

    Example of pysam header converted to dict:
    dict([
        ('SQ', [{'SN': 'chr1', 'LN': 248956422},
         {'SN': 'chr10', 'LN': 133797422},
         {'SN': 'chr11', 'LN': 135086622},
         {'SN': 'chr12', 'LN': 133275309}]),
        ('PG', [{'ID': 'bwa', 'PN': 'bwa', 'VN': '0.7.17-r1188', 'CL': 'bwa mem -t 8 -SP -v1 hg38.fa test_1.1.fastq.gz test_2.1.fastq.gz'}])
    ])
    """
    SQs = samheader.to_dict()["SQ"]
    chromsizes = [(sq["SN"], int(sq["LN"])) for sq in SQs]
    return dict(chromsizes)


def get_chrom_order(chroms_file, sam_chroms=None):
    """
    Produce an "enumeration" of chromosomes based on the list
    of chromosomes
    """
    chrom_enum = dict()
    i = 1
    with open(chroms_file, "rt") as f:
        for line in f:
            chrom = line.strip().split("\t")[0]
            if chrom and ((not sam_chroms) or (chrom in sam_chroms)):
                chrom_enum[chrom] = i
                i += 1

    if sam_chroms:
        remaining = sorted(
            chrom for chrom in sam_chroms if chrom not in chrom_enum.keys()
        )

        for chrom in remaining:
            chrom_enum[chrom] = i
            i += 1

    return chrom_enum


def make_standard_pairsheader(
    assembly=None,
    chromsizes=None,
    columns=pairsam_format.COLUMNS,
    shape="upper triangle",
):
    header = []
    header.append("## pairs format v{}".format(PAIRS_FORMAT_VERSION))
    header.append("#shape: {}".format(shape))

    header.append(
        "#genome_assembly: {}".format(assembly if assembly is not None else "unknown")
    )

    if chromsizes is not None:
        try:
            chromsizes = chromsizes.items()
        except AttributeError:
            pass
        for chrom, length in chromsizes:
            header.append("#chromsize: {} {}".format(chrom, length))

    header.append("#columns: " + SEP_COLS.join(columns))

    return header


def subset_chroms_in_pairsheader(header, chrom_subset):
    new_header = []
    for line in header:
        if line.startswith("#chromsize:"):
            if line.strip().split()[1] in chrom_subset:
                new_header.append(line)
        elif line.startswith("#chromosomes:"):
            line = SEP_CHROMS.join(
                ["#chromosomes:"]
                + [c for c in line.strip().split()[1:] if c in chrom_subset]
            )
            new_header.append(line)
        else:
            new_header.append(line)
    return new_header


def insert_samheader(header, samheader):
    """Insert samheader into header."""
    new_header = [l for l in header if not l.startswith("#columns")]
    if samheader:
        new_header += ["#samheader: " + l for l in samheader]
    new_header += [l for l in header if l.startswith("#columns")]
    return new_header


def insert_samheader_pysam(header, samheader):
    """Insert samheader into header,pysam version."""
    new_header = [l for l in header if not l.startswith("#columns")]
    if samheader:
        new_header += ["#samheader: " + l for l in str(samheader).strip().split("\n")]
    new_header += [l for l in header if l.startswith("#columns")]
    return new_header


def mark_header_as_sorted(header):
    header = copy.deepcopy(header)
    if is_empty_header(header):
        raise Exception("Input file is not valid .pairs, has no header or is empty.")
    if not any([l.startswith("#sorted") for l in header]):
        if header[0].startswith("##"):
            header.insert(1, "#sorted: chr1-chr2-pos1-pos2")
        else:
            header.insert(0, "#sorted: chr1-chr2-pos1-pos2")
    for i in range(len(header)):
        if header[i].startswith("#chromosomes"):
            chroms = header[i][12:].strip().split(SEP_CHROMS)
            header[i] = "#chromosomes: {}".format(SEP_CHROMS.join(sorted(chroms)))
    return header


def append_new_pg(header, ID="", PN="", VN=None, CL=None, force=False):
    header = copy.deepcopy(header)
    if is_empty_header(header):
        raise Exception("Input file is not valid .pairs, has no header or is empty.")
    samheader, other_header = extract_fields(header, "samheader", save_rest=True)
    new_samheader = _add_pg_to_samheader(samheader, ID, PN, VN, CL, force)
    new_header = insert_samheader(other_header, new_samheader)
    return new_header


def _update_header_entry(header, field, new_value):
    header = copy.deepcopy(header)
    found = False
    newline = "#{}: {}".format(field, new_value)
    for i in range(len(header)):
        if header[i].startswith(COMMENT_CHAR + field):
            header[i] = newline
            found = True
    if not found:
        if header[-1].startswith("#columns"):
            header.insert(-1, newline)
        else:
            header.append(newline)
    return header


def _add_pg_to_samheader(samheader, ID="", PN="", VN=None, CL=None, force=False):
    """Append a @PG record to an existing sam header. If the header comes
    from a merged file and thus has multiple chains of @PG, append the
    provided PG to all of the chains, adding the numerical suffix of the
    branch to the ID.

    Parameters
    ----------

    header : list of str
    ID, PN, VN, CL : std
        The keys of a new @PG record. If absent, VN is the version of
        pairtools and CL is taken from sys.argv.
    force : bool
        If True, ignore the inconsistencies among @PG records of the existing
        header.

    Returns
    -------
    new_header : list of str
        A list of new headers lines, stripped of newline characters.
    """

    if VN is None:
        VN = __version__
    if CL is None:
        CL = " ".join(sys.argv)

    pre_pg_header = [
        line.strip()
        for line in samheader
        if line.startswith("@HD") or line.startswith("@SQ") or line.startswith("@RG")
    ]

    post_pg_header = [
        line.strip()
        for line in samheader
        if not line.startswith("@HD")
        and (not line.startswith("@SQ"))
        and (not line.startswith("@RG"))
        and (not line.startswith("@PG"))
    ]

    pg_chains = _parse_pg_chains(samheader, force=force)

    for i, br in enumerate(pg_chains):
        new_pg = {"ID": ID, "PN": PN, "VN": VN, "CL": CL}
        new_pg["PP"] = br[-1]["ID"]
        if len(pg_chains) > 1:
            new_pg["ID"] = new_pg["ID"] + "-" + str(i + 1) + "." + str(len(br) + 1)
        new_pg["raw"] = _format_pg(**new_pg)
        br.append(new_pg)

    new_header = (
        pre_pg_header + [pg["raw"] for br in pg_chains for pg in br] + post_pg_header
    )

    return new_header


def _format_pg(**kwargs):
    out = ["@PG"] + [
        "{}:{}".format(field, kwargs[field])
        for field in ["ID", "PN", "CL", "PP", "DS", "VN"]
        if field in kwargs
    ]
    return "\t".join(out)


def _parse_pg_chains(header, force=False):
    pg_chains = []

    parsed_pgs = []
    for l in header:
        if l.startswith("@PG"):
            tag_value_pairs = l.strip().split("\t")[1:]
            if not all(":" in tvp for tvp in tag_value_pairs):
                warnings.warn(
                    f"Skipping the following @PG line, as it does not follow the SAM header standard of TAG:VALUE: {l}"
                )
                continue
            parsed_tvp = dict(
                [tvp.split(":", maxsplit=1) for tvp in tag_value_pairs if ":" in tvp]
            )
            if parsed_tvp:
                parsed_tvp["raw"] = l.strip()
                parsed_pgs.append(parsed_tvp)

    while True:
        if len(parsed_pgs) == 0:
            break

        for i in range(len(parsed_pgs)):
            pg = parsed_pgs[i]
            if "PP" not in pg:
                pg_chains.append([pg])
                parsed_pgs.pop(i)
                break
            else:
                matching_chains = [
                    branch for branch in pg_chains if branch[-1]["ID"] == pg["PP"]
                ]
                if len(matching_chains) > 1:
                    if force:
                        matching_chains[0].append(pg)
                        parsed_pgs.pop(i)
                        break

                    else:
                        raise ParseError(
                            "Multiple @PG records with the IDs identical to the PP field of another record:\n"
                            + "\n".join([br[-1]["raw"] for br in matching_chains])
                            + "\nvs\n"
                            + pg["raw"]
                        )

                if len(matching_chains) == 1:
                    matching_chains[0].append(pg)
                    parsed_pgs.pop(i)
                    break

            if force:
                pg_chains.append([pg])
                parsed_pgs.pop(i)
                break
            else:
                raise ParseError(
                    "Cannot find the parental @PG record for the @PG records:\n"
                    + "\n".join([pg["raw"] for pg in parsed_pgs])
                )

    return pg_chains


def _toposort(dag, tie_breaker):
    """
    Topological sort on a directed acyclic graph

    Uses Kahn's algorithm with a custom tie-breaking option. The
    dictionary ``dag`` can be interpreted in two ways:

    1. A dependency graph (i.e. arcs point from values to keys),
       and the generator yields items with no dependences followed
       by items that depend on previous ones.

    2. Arcs point from keys to values, in which case the generator produces
       a **reverse** topological ordering of the nodes.

    Parameters
    ----------
    dag: dict of nodes to sets of nodes
        Directed acyclic graph encoded as a dictionary.
    tie_breaker: callable
        Function that picks a tie breaker from a set of nodes with no
        unprocessed dependences.

    Returns
    -------
    Generator

    Notes
    -----
    See <https://en.wikipedia.org/wiki/Topological_sorting for more info>.
    Based in part on activestate recipe:
    <http://code.activestate.com/recipes/578272-topological-sort/> by Sam
    Denton (MIT licensed).

    """
    # Drop self-edges.
    for k, v in dag.items():
        v.discard(k)

    # Find all nodes that don't depend on anything
    # and include them with empty dependencies.
    indep_nodes = set.union(*dag.values()) - set(dag.keys())
    dag.update({node: set() for node in indep_nodes})

    while True:
        if not indep_nodes:
            break
        out = tie_breaker(indep_nodes)
        indep_nodes.discard(out)
        del dag[out]

        yield out

        for node, deps in dag.items():
            deps.discard(out)
            if len(deps) == 0:
                indep_nodes.add(node)

    if len(dag) != 0:
        raise ValueError("Circular dependencies exist: {} ".format(list(dag.items())))


def merge_chrom_lists(*lsts):
    sentinel = "!NONE!"

    g = defaultdict(set)
    for lst in lsts:
        if len(lst) == 1:
            g[lst[0]].add(sentinel)
        for a, b in zip(lst[:-1], lst[1:]):
            g[b].add(a)
    if len(g) == 0:
        return []

    chrom_list = list(_toposort(g.copy(), tie_breaker=min))
    if sentinel in chrom_list:
        chrom_list.remove(sentinel)
    chrom_list = sorted(chrom_list)
    return chrom_list


def _merge_samheaders(samheaders, force=False):
    # first, append an HD line if it is present in any files
    # if different lines are present, raise an error
    HDs = set.union(
        *[
            set(line for line in samheader if line.startswith("@HD"))
            for samheader in samheaders
        ]
    )
    if len(HDs) > 1 and not force:
        raise ParseError("More than one unique @HD line is found in samheaders!")
    HDs = [list(HDs)[0]] if HDs else []

    # second, confirm that all files had the same SQ lines
    # add SQs from the first file, keeping its order
    SQs = [
        set(line for line in samheader if line.startswith("@SQ"))
        for samheader in samheaders
    ]
    common_SQs = set.intersection(*SQs)

    SQs_same = all([len(samheader) == len(common_SQs) for samheader in SQs])

    if not SQs_same and not (force):
        raise ParseError("The SQ (sequence) lines of the sam headers are not identical")
    SQs = [line for line in samheaders[0] if line.startswith("@SQ")]

    # third, append _all_ PG chains, adding a unique index according to the
    # provided merging order
    PGs = []
    for i, samheader in enumerate(samheaders):
        for line in samheader:
            if line.startswith("@PG"):
                split_line = line.split("\t")
                for j in range(len(split_line)):
                    if split_line[j].startswith("ID:") or split_line[j].startswith(
                        "PP:"
                    ):
                        split_line[j] = split_line[j] + "-" + str(i + 1)
                PGs.append("\t".join(split_line))

    # finally, add all residual unique lines
    rest = sum(
        [
            list(
                set(
                    line
                    for line in samheader
                    if (not line.startswith("@HD"))
                    and (not line.startswith("@SQ"))
                    and (not line.startswith("@PG"))
                )
            )
            for samheader in samheaders
        ],
        [],
    )

    new_header = []
    new_header += HDs
    new_header += SQs
    new_header += PGs
    new_header += rest

    return new_header


def _merge_pairheaders(pairheaders, force=False):
    new_header = []

    # first, add all keys that are expected to be the same among all headers
    keys_expected_identical = [
        "## pairs format",
        "#sorted:",
        "#shape:",
        "#genome_assembly:",
        "#columns:",
    ]

    keys_orginal = [l.split()[0] for header in pairheaders for l in header]

    for k in keys_expected_identical:
        lines = [[l for l in header if l.startswith(k)] for header in pairheaders]
        same = all([l == lines[0] for l in lines])
        if not (same or force):
            raise ParseError(
                "The following header entries must be the same "
                "the merged files: {}".format(k)
            )
        new_header += lines[0]

    # second, merge and add the chromsizes fields.
    chrom_lists = []
    chromsizes = {}
    for header in pairheaders:
        chromlist = []
        for line in header:
            if line.startswith("#chromsize:"):
                chrom, length = line.strip("#chromsize:").split()
                chromsizes[chrom] = length
                chromlist.append(chrom)
        chrom_lists.append(chromlist)

    chroms_merged = merge_chrom_lists(*chrom_lists)
    if "#chromosomes:" in keys_orginal:
        chrom_line = "#chromosomes: {}".format(" ".join(chroms_merged))
        new_header.extend([chrom_line])

    chromsize_lines = [
        "#chromsize: {} {}".format(chrom, chromsizes[chrom]) for chrom in chroms_merged
    ]
    new_header.extend(chromsize_lines)

    # finally, add a sorted list of other unique fields
    other_lines = sorted(
        set(
            l
            for h in pairheaders
            for l in h
            if not any(
                l.startswith(k)
                for k in keys_expected_identical + ["#chromosomes", "#chromsize"]
            )
        )
    )
    if other_lines:
        if new_header[-1].startswith("#columns"):
            new_header = new_header[:-1] + other_lines + [new_header[-1]]
        else:
            new_header = new_header + other_lines

    return new_header


def all_same_columns(pairheaders):
    key_target = "#columns:"
    lines = [[l for l in header if l.startswith(key_target)] for header in pairheaders]
    all_same = all([l == lines[0] for l in lines])
    return all_same


def merge_headers(headers, force=False):
    samheaders, pairheaders = zip(
        *[extract_fields(h, "samheader", save_rest=True) for h in headers]
    )
    # HD headers contain information that becomes invalid after processing
    # with distiller. Do not print into the output.
    new_pairheader = _merge_pairheaders(pairheaders, force=False)
    new_samheader = _merge_samheaders(samheaders, force=force)

    new_header = insert_samheader(new_pairheader, new_samheader)

    return new_header


def append_columns(header, columns):
    """
    Appends columns to the header, separated by SEP_COLS

    Parameters
    ----------
    header: Previous header
    columns: List of column names to append

    Returns
    -------
    Modified header (appended columns to the field "#columns")
    """
    for i in range(len(header)):
        if header[i].startswith("#columns: "):
            header[i] += SEP_COLS + SEP_COLS.join(columns)
    return header


def get_colnames(header):
    """
    Get column names of the header, separated by SEP_COLS

    Parameters
    ----------
    header: Previous header

    Returns
    -------
    List of column names
    """
    for i in range(len(header)):
        if header[i].startswith("#columns: "):
            columns = header[i].split(SEP_COLS)[1:]
            return columns
    return []


def set_columns(header, columns):
    """
    Set columns to the header, separated by SEP_COLS

    Parameters
    ----------
    header: Previous header
    columns: List of column names to append

    Returns
    -------
    Modified header (appended columns to the field "#columns")
    """
    for i in range(len(header)):
        if header[i].startswith("#columns:"):
            header[i] = "#columns:" + SEP_COLS + SEP_COLS.join(columns)
    return header


# def _guess_genome_assembly(samheader):
#    PG = [l for l in samheader if l.startswith('@PG') and '\tID:bwa' in l][0]
#    CL = [field for field in PG.split('\t') if field.startswith('CL:')]
#
#    return ga



from . import fileio
from . import dedup
from . import dedup_cython
from . import filterbycov
from . import headerops
from . import pairsam_format
from . import parse
from . import parse_pysam
from . import restrict
from . import stats
from . import select



# -*- coding: utf-8 -*-
import os
import sys

import pytest

import subprocess

testdir = os.path.dirname(os.path.realpath(__file__))


def test_python_version():
    assert sys.version_info[0] == 3, "Use Python 3!"


def test_mock_pysam():
    """Parse non-chimeric alignments with walks-policy mask with pysam backend."""
    mock_sam_path = os.path.join(testdir, "data", "mock.sam")
    mock_chroms_path = os.path.join(testdir, "data", "mock.chrom.sizes")
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "parse",
                "--walks-policy",
                "mask",
                "-c",
                mock_chroms_path,
                mock_sam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check if the header got transferred correctly
    sam_header = [l.strip() for l in open(mock_sam_path, "r") if l.startswith("@")]
    pairsam_header = [l.strip() for l in result.split("\n") if l.startswith("#")]
    for l in sam_header:
        assert any([l in l2 for l2 in pairsam_header])

    # check that the pairs got assigned properly
    for l in result.split("\n"):
        if l.startswith("#") or not l:
            continue

        print(l)
        assigned_pair = l.split("\t")[1:8]
        simulated_pair = l.split("CT:Z:SIMULATED:", 1)[1].split("\031", 1)[0].split(",")
        print(assigned_pair)
        print(simulated_pair)
        print()

        assert assigned_pair == simulated_pair


def test_mock_pysam_parse_all():
    """Parse all alignment in each read with walks-policy all and pysam backend."""
    mock_sam_path = os.path.join(testdir, "data", "mock.parse-all.sam")
    mock_chroms_path = os.path.join(testdir, "data", "mock.chrom.sizes")
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "parse",
                "--walks-policy",
                "all",
                "-c",
                mock_chroms_path,
                "--add-pair-index",
                mock_sam_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check if the header got transferred correctly
    sam_header = [l.strip() for l in open(mock_sam_path, "r") if l.startswith("@")]
    pairsam_header = [l.strip() for l in result.split("\n") if l.startswith("#")]
    for l in sam_header:
        assert any([l in l2 for l2 in pairsam_header])

    # check that the pairs got assigned properly
    id_counter = 0
    prev_id = ""
    for l in result.split("\n"):
        if l.startswith("#") or not l:
            continue

        if prev_id == l.split("\t")[0]:
            id_counter += 1
        else:
            id_counter = 0
        prev_id = l.split("\t")[0]

        assigned_pair = l.split("\t")[1:8] + l.split("\t")[-2:]
        simulated_pair = (
            l.split("CT:Z:SIMULATED:", 1)[1]
            .split("\031", 1)[0]
            .split("|")[id_counter]
            .split(",")
        )
        print(assigned_pair)
        print(simulated_pair, prev_id)
        print()

        assert assigned_pair == simulated_pair



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest


import tempfile

testdir = os.path.dirname(os.path.realpath(__file__))

tmpdir = tempfile.TemporaryDirectory()
tmpdir_name = tmpdir.name
mock_pairsam_path_1 = os.path.join(testdir, "data", "mock.pairsam")
mock_pairsam_path_2 = os.path.join(testdir, "data", "mock.2.pairsam")
mock_sorted_pairsam_path_1 = os.path.join(tmpdir_name, "1.pairsam")
mock_sorted_pairsam_path_2 = os.path.join(tmpdir_name, "2.pairsam")


@pytest.fixture
def setup_sort_two():
    try:
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "sort",
                mock_pairsam_path_1,
                "--output",
                mock_sorted_pairsam_path_1,
            ],
        )

        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "sort",
                mock_pairsam_path_2,
                "--output",
                mock_sorted_pairsam_path_2,
            ],
        )
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e


def test_mock_pairsam(setup_sort_two):

    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "merge",
                mock_sorted_pairsam_path_1,
                mock_sorted_pairsam_path_2,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check that all pairsam entries survived sorting:
    pairsam_body_1 = [
        l.strip()
        for l in open(mock_pairsam_path_1, "r")
        if not l.startswith("#") and l.strip()
    ]
    pairsam_body_2 = [
        l.strip()
        for l in open(mock_pairsam_path_2, "r")
        if not l.startswith("#") and l.strip()
    ]
    output_body = [
        l.strip() for l in result.split("\n") if not l.startswith("#") and l.strip()
    ]
    assert len(pairsam_body_1) + len(pairsam_body_2) == len(output_body)

    # check the sorting order of the output:
    prev_pair = None
    for l in output_body:
        cur_pair = l.split("\t")[1:8]
        if prev_pair is not None:
            assert cur_pair[0] >= prev_pair[0]
            if cur_pair[0] == prev_pair[0]:
                assert cur_pair[1] >= prev_pair[1]
                if cur_pair[1] == prev_pair[1]:
                    assert cur_pair[2] >= prev_pair[2]
                    if cur_pair[2] == prev_pair[2]:
                        assert cur_pair[3] >= prev_pair[3]

        prev_pair = cur_pair

    # Check that the header is preserved:
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "merge",
                "--keep-first-header",
                mock_sorted_pairsam_path_1,
                mock_sorted_pairsam_path_2,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check the headers:
    pairsam_header_1 = [
        l.strip()
        for l in open(mock_sorted_pairsam_path_1, "r")
        if l.startswith("#") and l.strip()
    ]
    pairsam_header_2 = [
        l.strip()
        for l in open(mock_sorted_pairsam_path_2, "r")
        if l.startswith("#") and l.strip()
    ]
    output_header = [
        l.strip() for l in result.split("\n") if l.startswith("#") and l.strip()
    ]

    assert len(pairsam_header_1) + 1 == len(output_header)

    tmpdir.cleanup()



# -*- coding: utf-8 -*-
import os
import sys

import pytest

import subprocess

testdir = os.path.dirname(os.path.realpath(__file__))


def test_restrict():
    """Restrict pairs file"""
    mock_pairs_path = os.path.join(testdir, "data", "mock.test-restr.pairs")
    mock_rfrag_path = os.path.join(testdir, "data", "mock.rsites.bed")
    try:
        result = subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "restrict",
                "-f",
                mock_rfrag_path,
                mock_pairs_path,
            ],
        ).decode("ascii")
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e

    # check if the header got transferred correctly
    true_header = [l.strip() for l in open(mock_pairs_path, "r") if l.startswith("@")]
    output_header = [l.strip() for l in result.split("\n") if l.startswith("#")]
    for l in true_header:
        assert any([l in l2 for l2 in output_header])

    # check that the pairs got assigned properly
    cols = [x for x in output_header if x.startswith("#columns")][0].split(" ")[1:]

    COL_RFRAG1_TRUE = cols.index("rfrag_test1")
    COL_RFRAG2_TRUE = cols.index("rfrag_test2")
    COL_RFRAG1_OUTPUT = cols.index("rfrag1")
    COL_RFRAG2_OUTPUT = cols.index("rfrag2")

    for l in result.split("\n"):
        if l.startswith("#") or not l:
            continue

        line = l.split()
        assert line[COL_RFRAG1_TRUE] == line[COL_RFRAG1_OUTPUT]
        assert line[COL_RFRAG2_TRUE] == line[COL_RFRAG2_OUTPUT]



import sys
import click
import re, fnmatch

from ..lib import fileio, pairsam_format, headerops
from . import cli, common_io_options

from ..lib.phase import phase_side_XB, phase_side_XA

UTIL_NAME = "pairtools_phase"


@cli.command()
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@click.option(
    "--phase-suffixes",
    nargs=2,
    # type=click.Tuple([str, str]),
    help="Phase suffixes (of the chrom names), always a pair.",
)
@click.option(
    "--clean-output",
    is_flag=True,
    help="Drop all columns besides the standard ones and phase1/2",
)
@click.option(
    "--tag-mode",
    type=click.Choice(["XB", "XA"]),
    default="XB",
    help="Specifies the mode of bwa reporting."
    " XA will parse 'XA', the input should be generated with: --add-columns XA,NM,AS,XS --min-mapq 0"
    " XB will parse 'XB' tag, the input should be generated with: --add-columns XB,AS,XS --min-mapq 0 "
    " Note that XB tag is added by running bwa with -u tag, present in github version. "
    " Both modes report similar results: XB reports 0.002% contacts more for phased data, "
    " while XA can report ~1-2% more unphased contacts because its definition multiple mappers is more premissive. ",
)
@click.option(
    "--report-scores/--no-report-scores",
    is_flag=True,
    default=False,
    help="Report scores of optional, suboptimal and second suboptimal alignments. "
    "NM (edit distance) with --tag-mode XA and AS (alfn score) with --tag-mode XB ",
)
@common_io_options
def phase(
    pairs_path, output, phase_suffixes, clean_output, tag_mode, report_scores, **kwargs
):
    """Phase pairs mapped to a diploid genome.
    Diploid genome is the genome with two set of the chromosome variants,
    where each chromosome has one of two suffixes (phase-suffixes)
    corresponding to the genome version (phase-suffixes).

    By default, phasing adds two additional columns with phase 0, 1 or "." (unpahsed).

    Phasing is based on detection of chromosome origin of each mapped fragment.
    Three scores are considered: best alignment score (S1),
    suboptimal alignment (S2) and second suboptimal alignment (S3) scores.
    Each fragment can be:
    1) uniquely mapped and phased (S1>S2>S3, first alignment is the best hit),
    2) uniquely mapped but unphased (S1=S2>S3, cannot distinguish between chromosome variants),
    3) multiply mapped (S1=S2=S3) or unmapped.

    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip/lz4c. By default, the input is read from stdin.

    """
    phase_py(
        pairs_path,
        output,
        phase_suffixes,
        clean_output,
        tag_mode,
        report_scores,
        **kwargs
    )


if __name__ == "__main__":
    phase()


def phase_py(
    pairs_path, output, phase_suffixes, clean_output, tag_mode, report_scores, **kwargs
):

    instream = (
        fileio.auto_open(
            pairs_path,
            mode="r",
            nproc=kwargs.get("nproc_in"),
            command=kwargs.get("cmd_in", None),
        )
        if pairs_path
        else sys.stdin
    )
    outstream = (
        fileio.auto_open(
            output,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )
        if output
        else sys.stdout
    )

    header, body_stream = headerops.get_header(instream)
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)
    old_column_names = headerops.extract_column_names(header)

    idx_phase1 = len(old_column_names)
    idx_phase2 = len(old_column_names) + 1
    if clean_output:
        new_column_names = [
            col for col in old_column_names if col in pairsam_format.COLUMNS
        ]
        new_column_idxs = [
            i for i, col in enumerate(old_column_names) if col in pairsam_format.COLUMNS
        ]
        new_column_idxs += [idx_phase1, idx_phase2]
    else:
        new_column_names = list(old_column_names)

    new_column_names.append("phase1")
    new_column_names.append("phase2")

    if report_scores:
        if tag_mode == "XB":
            new_column_names.append("S1_1")
            new_column_names.append("S1_2")
            new_column_names.append("S2_1")
            new_column_names.append("S2_2")
            new_column_names.append("S3_1")
            new_column_names.append("S3_2")
            if clean_output:
                new_column_idxs += [(idx_phase2 + i + 1) for i in range(6)]
        elif tag_mode == "XA":
            new_column_names.append("M1_1")
            new_column_names.append("M1_2")
            new_column_names.append("M2_1")
            new_column_names.append("M2_2")
            new_column_names.append("M3_1")
            new_column_names.append("M3_2")
            if clean_output:
                new_column_idxs += [(idx_phase2 + i + 1) for i in range(6)]
    header = headerops._update_header_entry(
        header, "columns", " ".join(new_column_names)
    )

    if tag_mode == "XB":
        if (
            ("XB1" not in old_column_names)
            or ("XB2" not in old_column_names)
            or ("AS1" not in old_column_names)
            or ("AS2" not in old_column_names)
            or ("XS1" not in old_column_names)
            or ("XS2" not in old_column_names)
        ):
            raise ValueError(
                "The input pairs file must be parsed with the flag --add-columns XB,AS,XS --min-mapq 0"
            )

        COL_XB1 = old_column_names.index("XB1")
        COL_XB2 = old_column_names.index("XB2")
        COL_AS1 = old_column_names.index("AS1")
        COL_AS2 = old_column_names.index("AS2")
        COL_XS1 = old_column_names.index("XS1")
        COL_XS2 = old_column_names.index("XS2")

    elif tag_mode == "XA":
        if (
            ("XA1" not in old_column_names)
            or ("XA2" not in old_column_names)
            or ("NM1" not in old_column_names)
            or ("NM2" not in old_column_names)
            or ("AS1" not in old_column_names)
            or ("AS2" not in old_column_names)
            or ("XS1" not in old_column_names)
            or ("XS2" not in old_column_names)
        ):
            raise ValueError(
                "The input pairs file must be parsed with the flag --add-columns XA,NM,AS,XS --min-mapq 0"
            )

        COL_XA1 = old_column_names.index("XA1")
        COL_XA2 = old_column_names.index("XA2")
        COL_NM1 = old_column_names.index("NM1")
        COL_NM2 = old_column_names.index("NM2")
        COL_AS1 = old_column_names.index("AS1")
        COL_AS2 = old_column_names.index("AS2")
        COL_XS1 = old_column_names.index("XS1")
        COL_XS2 = old_column_names.index("XS2")

    outstream.writelines((l + "\n" for l in header))

    for line in body_stream:
        cols = line.rstrip('\n').split(pairsam_format.PAIRSAM_SEP)
        cols.append("!")
        cols.append("!")
        if report_scores:
            for _ in range(6):
                cols.append("-1")
        pair_type = cols[pairsam_format.COL_PTYPE]

        if cols[pairsam_format.COL_C1] != pairsam_format.UNMAPPED_CHROM:
            if tag_mode == "XB":
                phase1, chrom_base1, S1_1, S2_1, S3_1 = phase_side_XB(
                    cols[pairsam_format.COL_C1],
                    cols[COL_XB1],
                    int(cols[COL_AS1]),
                    int(cols[COL_XS1]),
                    phase_suffixes,
                )
            elif tag_mode == "XA":
                phase1, chrom_base1, S1_1, S2_1, S3_1 = phase_side_XA(
                    cols[pairsam_format.COL_C1],
                    cols[COL_XA1],
                    int(cols[COL_AS1]),
                    int(cols[COL_XS1]),
                    int(cols[COL_NM1]),
                    phase_suffixes,
                )

            if not report_scores:
                cols[idx_phase1] = phase1
            else:
                (
                    cols[idx_phase1],
                    cols[idx_phase1 + 2],
                    cols[idx_phase1 + 4],
                    cols[idx_phase1 + 6],
                ) = (phase1, str(S1_1), str(S2_1), str(S3_1))
            cols[pairsam_format.COL_C1] = chrom_base1

            if chrom_base1 == "!":
                cols[pairsam_format.COL_C1] = pairsam_format.UNMAPPED_CHROM
                cols[pairsam_format.COL_P1] = str(pairsam_format.UNMAPPED_POS)
                cols[pairsam_format.COL_S1] = pairsam_format.UNMAPPED_STRAND
                pair_type = "M" + pair_type[1]

        if cols[pairsam_format.COL_C2] != pairsam_format.UNMAPPED_CHROM:

            if tag_mode == "XB":
                phase2, chrom_base2, S1_2, S2_2, S3_2 = phase_side_XB(
                    cols[pairsam_format.COL_C2],
                    cols[COL_XB2],
                    int(cols[COL_AS2]),
                    int(cols[COL_XS2]),
                    phase_suffixes,
                )
            elif tag_mode == "XA":
                phase2, chrom_base2, S1_2, S2_2, S3_2 = phase_side_XA(
                    cols[pairsam_format.COL_C2],
                    cols[COL_XA2],
                    int(cols[COL_AS2]),
                    int(cols[COL_XS2]),
                    int(cols[COL_NM2]),
                    phase_suffixes,
                )
            if not report_scores:
                cols[idx_phase2] = phase2
            else:
                (
                    cols[idx_phase2],
                    cols[idx_phase2 + 2],
                    cols[idx_phase2 + 4],
                    cols[idx_phase2 + 6],
                ) = (phase2, str(S1_2), str(S2_2), str(S3_2))
            cols[pairsam_format.COL_C2] = chrom_base2

            if chrom_base2 == "!":
                cols[pairsam_format.COL_C2] = pairsam_format.UNMAPPED_CHROM
                cols[pairsam_format.COL_P2] = str(pairsam_format.UNMAPPED_POS)
                cols[pairsam_format.COL_S2] = pairsam_format.UNMAPPED_STRAND
                pair_type = pair_type[0] + "M"

        cols[pairsam_format.COL_PTYPE] = pair_type

        if clean_output:
            cols = [cols[i] for i in new_column_idxs]

        outstream.write(pairsam_format.PAIRSAM_SEP.join(cols))
        outstream.write("\n")

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()


if __name__ == "__main__":
    phase()



import shutil
import pipes
import subprocess
import sys


class ParseError(Exception):
    pass


def auto_open(path, mode, nproc=1, command=None):
    """Guess the file format from the extension and use the corresponding binary
    to open it for reading or writing. If the extension is not known, open the
    file as text.

    If the binary allows parallel execution, specify the number of threads
    with `nproc`.

    If `command` is supplied, use it to open the file instead of auto-guessing.
    The command must accept the filename as the last argument, accept input
    through stdin and print output into stdout.

    Supported extensions and binaries (with comments):
    .bam - samtools view (allows parallel writing)
    .gz - pbgzip if available, otherwise bgzip
    .lz4 - lz4c (does not support parallel execution)
    """

    # Empty filepath or False provided
    if not path or path == "-":
        if mode == "r":
            return sys.stdin
        if mode == "w":
            return sys.stdout

    if command:
        if mode == "w":
            t = pipes.Template()
            t.append(command, "--")
            f = t.open(path, "w")
        elif mode == "r":
            t = pipes.Template()
            t.append(command, "--")
            f = t.open(path, "r")
        else:
            raise ValueError("Unknown mode : {}".format(mode))
        return f

    elif path.endswith(".bam"):
        if shutil.which("samtools") is None:
            raise ValueError(
                {
                    "w": "samtools is not found, cannot compress output",
                    "r": "samtools is not found, cannot decompress input",
                }[mode]
            )
        if mode == "w":
            t = pipes.Template()
            t.append(
                "samtools view -bS {} -".format(
                    "-@ " + str(nproc - 1) if nproc > 1 else ""
                ),
                "--",
            )
            f = t.open(path, "w")
        elif mode == "r":
            t = pipes.Template()
            t.append("samtools view -h", "--")
            f = t.open(path, "r")
        else:
            raise ValueError("Unknown mode for .bam : {}".format(mode))
        return f

    elif path.endswith(".gz"):
        if shutil.which("pbgzip") is not None:
            if mode == "w":
                t = pipes.Template()
                t.append("pbgzip -c -n {}".format(nproc), "--")
                f = t.open(path, "w")
            elif mode == "a":
                t = pipes.Template()
                t.append("pbgzip -c -n {} $IN >> $OUT".format(nproc), "ff")
                f = t.open(path, "w")
            elif mode == "r":
                t = pipes.Template()
                t.append("pbgzip -dc -n {}".format(nproc), "--")
                f = t.open(path, "r")
            else:
                raise ValueError("Unknown mode for .gz : {}".format(mode))
        elif shutil.which("bgzip") is not None:
            if mode == "w":
                t = pipes.Template()
                t.append("bgzip -c -@ {}".format(nproc), "--")
                f = t.open(path, "w")
            elif mode == "a":
                t = pipes.Template()
                t.append("bgzip -c -@ {} $IN >> $OUT".format(nproc), "ff")
                f = t.open(path, "w")
            elif mode == "r":
                t = pipes.Template()
                t.append("bgzip -dc -@ {}".format(nproc), "--")
                f = t.open(path, "r")
            else:
                raise ValueError("Unknown mode for .gz : {}".format(mode))
        elif shutil.which("gzip") is not None:
            if mode == "w":
                t = pipes.Template()
                t.append("gzip -c", "--")
                f = t.open(path, "w")
            elif mode == "a":
                t = pipes.Template()
                t.append("gzip -c $IN >> $OUT", "ff")
                f = t.open(path, "w")
            elif mode == "r":
                t = pipes.Template()
                t.append("gzip -dc", "--")
                f = t.open(path, "r")
            else:
                raise ValueError("Unknown mode for .gz : {}".format(mode))
        else:
            raise ValueError(
                {
                    "w": "pbgzip, bgzip and gzip are not found, cannot compress output",
                    "a": "pbgzip, bgzip and gzip are is not found, cannot compress output",
                    "r": "pbgzip, bgzip and gzip are is not found, cannot decompress input",
                }[mode]
            )
        return f
    elif path.endswith(".lz4"):
        if shutil.which("lz4c") is None:
            raise ValueError(
                {
                    "w": "lz4c is not found, cannot compress output",
                    "a": "lz4c is not found, cannot compress output",
                    "r": "lz4c is not found, cannot decompress input",
                }[mode]
            )
        if mode == "w":
            t = pipes.Template()
            t.append("lz4c -cz", "--")
            f = t.open(path, "w")
        elif mode == "a":
            t = pipes.Template()
            t.append("lz4c -cz $IN >> $OUT", "ff")
            f = t.open(path, "w")
        elif mode == "r":
            t = pipes.Template()
            t.append("lz4c -cd", "--")
            f = t.open(path, "r")
        else:
            raise ValueError("Unknown mode : {}".format(mode))
        return f
    else:
        return open(path, mode)


class PipedIO:
    def __init__(self, file_or_path, command, mode="r"):
        """
        An experimental class that reads/writes a file, piping the contents
        through another process.

        Parameters
        ----------
        file_or_path : file-like object or str
            A path to the input/output file or an already opened
            file-like object.
        command : str
            A command to launch a reading/writing process.
            If mode is 'w', the process must accept input via stdin.
            If mode is 'r', the process must put output into stdout.
            If mode is 'r' and file_or_path is str, the path will be
            appended to the command as the last argument.
        mode : str
            The mode for opening, same as in open(mode=).

        Returns
        -------
        file: a file-like object
        """

        if issubclass(type(command), str):
            command = command.split(" ")
        self._command = command
        self._mode = mode

        if mode.startswith("r"):
            if issubclass(type(file_or_path), str):
                self._proc = subprocess.Popen(
                    command + [file_or_path],
                    universal_newlines=True,
                    stdout=subprocess.PIPE,
                )
            else:
                self._proc = subprocess.Popen(
                    command,
                    universal_newlines=True,
                    stdin=file_or_path,
                    stdout=subprocess.PIPE,
                )
            self._stream = self._proc.stdout

            self._close_stream = self._proc.stdout.close

        elif mode.startswith("w") or mode.startswith("a"):
            f = (
                open(file_or_path, mode=mode)
                if issubclass(type(file_or_path), str)
                else file_or_path
            )
            self._proc = subprocess.Popen(
                command, universal_newlines=True, stdin=subprocess.PIPE, stdout=f
            )
            self._stream = self._proc.stdin

        self.buffer = self._stream.buffer
        self.closed = self._stream.closed
        self.flush = self._stream.flush
        self.fileno = self._stream.fileno

        self.read = self._stream.read
        self.readline = self._stream.readline
        self.readlines = self._stream.readlines

        self.seek = self._stream.seek
        self.seekable = self._stream.seekable
        self.truncate = self._stream.truncate
        self.tell = self._stream.tell

        self.writable = self._stream.writable
        self.write = self._stream.write
        self.writelines = self._stream.writelines

    def close(self, timeout=None):
        self._stream.close()
        retcode = self._proc.wait(timeout=timeout)
        return retcode



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest
import tempfile

testdir = os.path.dirname(os.path.realpath(__file__))

tmpdir = tempfile.TemporaryDirectory()
tmpdir_name = tmpdir.name

mock_pairsam_path_dedup = os.path.join(testdir, "data", "mock.4dedup.pairsam")

dedup_path = os.path.join(tmpdir_name, "dedup.pairsam")
unmapped_path = os.path.join(tmpdir_name, "unmapped.pairsam")
dups_path = os.path.join(tmpdir_name, "dups.pairsam")

dedup_max_path = os.path.join(tmpdir_name, "dedup_max.pairsam")
unmapped_max_path = os.path.join(tmpdir_name, "unmapped_max.pairsam")
dups_max_path = os.path.join(tmpdir_name, "dups_max.pairsam")

dedup_markdups_path = os.path.join(tmpdir_name, "dedup.markdups.pairsam")
unmapped_markdups_path = os.path.join(tmpdir_name, "unmapped.markdups.pairsam")
dups_markdups_path = os.path.join(tmpdir_name, "dups.markdups.pairsam")

max_mismatch = 1


@pytest.fixture
def setup_dedup():
    try:
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "dedup",
                mock_pairsam_path_dedup,
                "--output",
                dedup_path,
                "--output-dups",
                dups_path,
                "--output-unmapped",
                unmapped_path,
                "--max-mismatch",
                str(max_mismatch),
            ],
        )
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "dedup",
                mock_pairsam_path_dedup,
                "--output",
                dedup_max_path,
                "--output-dups",
                dups_max_path,
                "--output-unmapped",
                unmapped_max_path,
                "--max-mismatch",
                str(max_mismatch),
                "--method",
                "max",
            ],
        )
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "dedup",
                mock_pairsam_path_dedup,
                "--mark-dups",
                "--output",
                dedup_markdups_path,
                "--output-dups",
                dups_markdups_path,
                "--output-unmapped",
                unmapped_markdups_path,
                "--max-mismatch",
                str(max_mismatch),
            ],
        )
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e


def test_mock_pairsam(setup_dedup):

    pairsam_pairs = [
        l.strip().split("\t")
        for l in open(mock_pairsam_path_dedup, "r")
        if not l.startswith("#") and l.strip()
    ]
    for (ddp, up, dp) in [
        (dedup_path, unmapped_path, dups_path),
        (dedup_max_path, unmapped_max_path, dups_max_path),
        (dedup_markdups_path, unmapped_markdups_path, dups_markdups_path),
    ]:

        dedup_pairs = [
            l.strip().split("\t")
            for l in open(ddp, "r")
            if not l.startswith("#") and l.strip()
        ]
        unmapped_pairs = [
            l.strip().split("\t")
            for l in open(up, "r")
            if not l.startswith("#") and l.strip()
        ]
        dup_pairs = [
            l.strip().split("\t")
            for l in open(dp, "r")
            if not l.startswith("#") and l.strip()
        ]

        # check that at least a few pairs remained in deduped and dup files
        assert len(dedup_pairs) > 0
        assert len(dup_pairs) > 0
        assert len(unmapped_pairs) > 0
        import pandas as pd

        # check that all pairsam entries survived deduping:

        assert len(dedup_pairs) + len(unmapped_pairs) + len(dup_pairs) == len(
            pairsam_pairs
        )

        def pairs_overlap(pair1, pair2, max_mismatch):
            overlap = (
                (pair1[1] == pair2[1])
                and (pair1[3] == pair2[3])
                and (pair1[5] == pair2[5])
                and (pair1[6] == pair2[6])
                and (abs(int(pair1[2]) - int(pair2[2])) <= max_mismatch)
                and (abs(int(pair1[4]) - int(pair2[4])) <= max_mismatch)
            )
            return overlap

        # check that deduped pairs do not overlap
        assert all(
            [
                not pairs_overlap(pair1, pair2, max_mismatch)
                for i, pair1 in enumerate(dedup_pairs)
                for j, pair2 in enumerate(dedup_pairs)
                if i != j
            ]
        )

        # check that the removed duplicates overlap with at least one of the
        # deduplicated entries
        assert all(
            [
                any([pairs_overlap(pair1, pair2, 3) for pair2 in dedup_pairs])
                for pair1 in dup_pairs
            ]
        )

    tmpdir.cleanup()



from . import pairsam_format
import warnings


def find_rfrag(rfrags, chrom, pos):

    # Return empty if chromosome is unmapped:
    if chrom == pairsam_format.UNMAPPED_CHROM:
        return (
            pairsam_format.UNANNOTATED_RFRAG,
            pairsam_format.UNMAPPED_POS,
            pairsam_format.UNMAPPED_POS,
        )

    try:
        rsites_chrom = rfrags[chrom]
    except ValueError as e:
        warnings.warn(
            f"Chomosome {chrom} does not have annotated restriction fragments, return empty."
        )
        return (
            pairsam_format.UNANNOTATED_RFRAG,
            pairsam_format.UNMAPPED_POS,
            pairsam_format.UNMAPPED_POS,
        )

    idx = min(
        max(0, rsites_chrom.searchsorted(pos, "right") - 1), len(rsites_chrom) - 2
    )
    return idx, rsites_chrom[idx], rsites_chrom[idx + 1]



import sys
import click
import re, fnmatch
import warnings

from ..lib import fileio, pairsam_format, headerops
from ..lib.select import evaluate_stream
from . import cli, common_io_options

UTIL_NAME = "pairtools_select"


@cli.command()
@click.argument("condition", type=str)
@click.argument("pairs_path", type=str, required=False)
@click.option(
    "-o",
    "--output",
    type=str,
    default="",
    help="output file."
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, the output is printed into stdout.",
)
@click.option(
    "--output-rest",
    type=str,
    default="",
    help="output file for pairs of other types. "
    " If the path ends with .gz or .lz4, the output is bgzip-/lz4c-compressed."
    " By default, such pairs are dropped.",
)

# Deprecated option to be removed in the future:
# @click.option(
#     "--send-comments-to",
#     type=click.Choice(['selected', 'rest', 'both', 'none']),
#     default="both",
#     help="Which of the outputs should receive header and comment lines",
#     show_default=True)


@click.option(
    "--chrom-subset",
    type=str,
    default=None,
    help="A path to a chromosomes file (tab-separated, 1st column contains "
    "chromosome names) containing a chromosome subset of interest. "
    "If provided, additionally filter pairs with both sides originating from "
    "the provided subset of chromosomes. This operation modifies the #chromosomes: "
    "and #chromsize: header fields accordingly.",
)
@click.option(
    "--startup-code",
    type=str,
    default=None,
    help="An auxiliary code to execute before filtering. "
    "Use to define functions that can be evaluated in the CONDITION statement",
)
@click.option(
    "-t",
    "--type-cast",
    type=(str, str),
    default=(),
    multiple=True,
    help="Cast a given column to a given type. By default, only pos and mapq "
    "are cast to int, other columns are kept as str. Provide as "
    "-t <column_name> <type>, e.g. -t read_len1 int. Multiple entries are allowed.",
)
@click.option(
    "--remove-columns",
    "-r",
    help=f"Comma-separated list of columns to be removed, e.g.: {','.join(pairsam_format.COLUMNS)}",
    type=str,
    default="",
    required=False,
)
@common_io_options
def select(
    condition,
    pairs_path,
    output,
    output_rest,  # send_comments_to,
    chrom_subset,
    startup_code,
    type_cast,
    remove_columns,
    **kwargs,
):
    """Select pairs according to some condition.

    CONDITION : A Python expression; if it returns True, select the read pair.
    Any column declared in the #columns line of the pairs header can be
    accessed by its name. If the header lacks the #columns line, the columns
    are assumed to follow the .pairs/.pairsam standard (readID, chrom1, chrom2,
    pos1, pos2, strand1, strand2, pair_type). Finally, CONDITION has access to
    COLS list which contains the string values of columns. In Bash, quote
    CONDITION with single quotes, and use double quotes for string variables
    inside CONDITION.

    PAIRS_PATH : input .pairs/.pairsam file. If the path ends with .gz or .lz4, the
    input is decompressed by bgzip/lz4c. By default, the input is read from stdin.

    The following functions can be used in CONDITION besides the standard Python functions:

    - csv_match(x, csv) - True if variable x is contained in a list of
    comma-separated values, e.g. csv_match(chrom1, 'chr1,chr2')

    - wildcard_match(x, wildcard) - True if variable x matches a wildcard,
    e.g. wildcard_match(pair_type, 'C*')

    - regex_match(x, regex) - True if variable x matches a Python-flavor regex,
    e.g. regex_match(chrom1, 'chr\d')

    \b
    Examples:
    pairtools select '(pair_type=="UU") or (pair_type=="UR") or (pair_type=="RU")'
    pairtools select 'chrom1==chrom2'
    pairtools select 'COLS[1]==COLS[3]'
    pairtools select '(chrom1==chrom2) and (abs(pos1 - pos2) < 1e6)'
    pairtools select '(chrom1=="!") and (chrom2!="!")'
    pairtools select 'regex_match(chrom1, "chr\d+") and regex_match(chrom2, "chr\d+")'

    pairtools select 'True' --chrom-subset mm9.reduced.chromsizes

    """
    select_py(
        condition,
        pairs_path,
        output,
        output_rest,  # send_comments_to,
        chrom_subset,
        startup_code,
        type_cast,
        remove_columns,
        **kwargs,
    )


def select_py(
    condition,
    pairs_path,
    output,
    output_rest,  # send_comments_to,
    chrom_subset,
    startup_code,
    type_cast,
    remove_columns,
    **kwargs,
):

    instream = fileio.auto_open(
        pairs_path,
        mode="r",
        nproc=kwargs.get("nproc_in"),
        command=kwargs.get("cmd_in", None),
    )
    outstream = fileio.auto_open(
        output,
        mode="w",
        nproc=kwargs.get("nproc_out"),
        command=kwargs.get("cmd_out", None),
    )

    # Optional output created only if requested:
    outstream_rest = None
    if output_rest:
        outstream_rest = fileio.auto_open(
            output_rest,
            mode="w",
            nproc=kwargs.get("nproc_out"),
            command=kwargs.get("cmd_out", None),
        )

    # Parse the input stream:
    header, body_stream = headerops.get_header(instream)

    # Modify the header:
    header = headerops.append_new_pg(header, ID=UTIL_NAME, PN=UTIL_NAME)

    # Filter out unwanted columns:
    if remove_columns:
        input_columns = headerops.extract_column_names(header)
        remove_columns = remove_columns.split(",")
        for col in remove_columns:
            if col in pairsam_format.COLUMNS_PAIRS:
                warnings.warn(
                    f"Removing required {col} column for .pairs format. Output is not .pairs anymore"
                )
            elif col in pairsam_format.COLUMNS_PAIRSAM:
                warnings.warn(
                    f"Removing required {col} column for .pairsam format. Output is not .pairsam anymore"
                )
        updated_columns = [x for x in input_columns if x not in remove_columns]

        if len(updated_columns) == len(input_columns):
            warnings.warn(
                f"Some column(s) {','.join(remove_columns)} not in the file, the operation has no effect"
            )
        else:
            header = headerops.set_columns(header, updated_columns)

    # Update the chromosomes:
    new_chroms = None
    if chrom_subset is not None:
        new_chroms = [l.strip().split("\t")[0] for l in open(chrom_subset, "r")]

    if new_chroms is not None:
        header = headerops.subset_chroms_in_pairsheader(header, new_chroms)
    outstream.writelines((l + "\n" for l in header))
    if output_rest:
        outstream_rest.writelines((l + "\n" for l in header))

    column_names = headerops.extract_column_names(header)
    if len(column_names) == 0:
        column_names = pairsam_format.COLUMNS

    # Columns filtration rule:
    if remove_columns:
        column_scheme = [input_columns.index(COL) for COL in updated_columns]

    # Format the condition:
    condition = condition.strip()
    if new_chroms is not None:
        condition = (
            f"({condition}) and (chrom1 in {new_chroms}) and (chrom2 in {new_chroms})"
        )

    for filter_passed, line in evaluate_stream(
        body_stream, condition, column_names, type_cast, startup_code
    ):
        COLS = line.rstrip('\n').split(pairsam_format.PAIRSAM_SEP)

        if remove_columns:
            COLS = [
                COLS[idx] for idx in column_scheme
            ]  # re-order the columns according to the scheme:
            line = pairsam_format.PAIRSAM_SEP.join(COLS) + "\n"  # form the line

        if filter_passed:
            outstream.write(line)
        elif outstream_rest:
            outstream_rest.write(line)

    if instream != sys.stdin:
        instream.close()

    if outstream != sys.stdout:
        outstream.close()

    if output_rest and outstream_rest != sys.stdout:
        outstream_rest.close()


if __name__ == "__main__":
    select()



import numpy as np
import pandas as pd

from .regions import assign_regs_c
import bioframe


def geomprog(factor, start=1):
    yield start
    while True:
        start *= factor
        yield start


def _geomrange(start, end, factor, endpoint):
    prev = np.nan
    for i in geomprog(factor, start):
        x = int(round(i))

        if x > end:
            break

        if x == prev:
            continue

        prev = x
        yield x

    if endpoint and prev != end:
        yield end


def geomrange(start, end, factor, endpoint=False):
    return np.fromiter(_geomrange(start, end, factor, endpoint), dtype=int)


def geomspace(start, end, num=50, endpoint=True):
    factor = (end / start) ** (1 / num)
    return geomrange(start, end, factor, endpoint=endpoint)


def _to_float(arr_or_scalar):
    if np.isscalar(arr_or_scalar):
        return float(arr_or_scalar)
    else:
        return np.asarray(arr_or_scalar).astype(float)


def assign_regs(chroms, pos, regs):
    gb_regs = regs.sort_values(["chrom", "start", "end"]).groupby(["chrom"])

    regs_dict = {
        chrom.encode(): regs_per_chrom[["start", "end"]]
        .values.flatten()
        .astype(np.int64)
        for chrom, regs_per_chrom in gb_regs
    }

    return assign_regs_c(np.asarray(chroms).astype("bytes"), np.asarray(pos), regs_dict)


def cartesian_df_product(df1, df2, suffixes=["1", "2"]):
    return pd.merge(
        left=df1.assign(cartesian_product_dummy=1),
        right=df2.assign(cartesian_product_dummy=1),
        on=["cartesian_product_dummy"],
        how="outer",
        suffixes=suffixes,
    ).drop("cartesian_product_dummy", axis="columns")


def make_empty_scaling(regions, dist_bins, multiindex=True):

    if dist_bins[0] != 0:
        dist_bins = np.r_[0, dist_bins]
    if dist_bins[-1] != np.iinfo(np.int64).max:
        dist_bins = np.r_[dist_bins, np.iinfo(np.int64).max]

    strands_table = pd.DataFrame(
        {"strand1": ["+", "+", "-", "-"], "strand2": ["+", "-", "+", "-"]}
    )
    dists_table = pd.DataFrame(
        list(zip(dist_bins[:-1], dist_bins[1:])), columns=["min_dist", "max_dist"]
    )

    out = regions.join(regions, on=None, lsuffix="1", rsuffix="2")
    out = cartesian_df_product(out, strands_table)
    out = cartesian_df_product(out, dists_table)

    if multiindex:
        index_by = [
            "chrom1",
            "start1",
            "end1",
            "chrom2",
            "start2",
            "end2",
            "strand1",
            "strand2",
            "min_dist",
            "max_dist",
        ]
        out.set_index(index_by, inplace=True)

    return out


def make_empty_cross_region_table(
    regions, drop_same_reg=True, split_by_strand=True, multiindex=True
):
    out = cartesian_df_product(regions, regions)
    if split_by_strand:
        strands_table = pd.DataFrame(
            {"strand1": ["+", "+", "-", "-"], "strand2": ["+", "-", "+", "-"]}
        )
        out = cartesian_df_product(out, strands_table)

    if drop_same_reg:
        out = out[
            (out["chrom1"] != out["chrom2"])
            | (out["start1"] != out["start2"])
            | (out["end1"] != out["end2"])
        ]

    if multiindex:
        index_by = ["chrom1", "start1", "end1", "chrom2", "start2", "end2"]
        if split_by_strand:
            index_by += ["strand1", "strand2"]

        out.set_index(index_by, inplace=True)

    return out


def bins_pairs_by_distance(
    pairs_df, dist_bins, regions=None, chromsizes=None, ignore_trans=False
):

    dist_bins = np.r_[dist_bins, np.iinfo(np.int64).max]
    if regions is None:
        if chromsizes is None:
            chroms = sorted(
                set.union(set(pairs_df.chrom1.unique()), set(pairs_df.chrom2.unique()))
            )
            regions = pd.DataFrame({"chrom": chroms, "start": 0, "end": -1})
            regions = regions[["chrom", "start", "end"]]

            region_starts1, region_starts2 = 0, 0
            region_ends1, region_ends2 = -1, -1

        else:
            region_starts1, region_starts2 = 0, 0
            region_ends1 = pairs_df.chrom1.map(chromsizes).fillna(1).astype(np.int64)
            region_ends2 = pairs_df.chrom2.map(chromsizes).fillna(1).astype(np.int64)
            regions = pd.DataFrame(
                [
                    {"chrom": chrom, "start": 0, "end": length}
                    for chrom, length in chromsizes.items()
                ]
            )
            regions = regions[["chrom", "start", "end"]]

        try:
            regions = bioframe.from_any(regions)
        except Exception as e:
            raise ValueError(f"Invalid viewframe created from pairs file, {e}")

    else:

        if not bioframe.is_viewframe(regions):
            try:
                regions = bioframe.from_any(regions)
            except Exception as e:
                raise ValueError(
                    f"Provided regions cannot be converted to viewframe, {e}"
                )

        regions = regions[["chrom", "start", "end"]]

        _, region_starts1, region_ends1 = assign_regs(
            pairs_df.chrom1.values, pairs_df.pos1.values, regions
        ).T
        _, region_starts2, region_ends2 = assign_regs(
            pairs_df.chrom2.values, pairs_df.pos2.values, regions
        ).T

    pairs_reduced_df = pd.DataFrame(
        {
            "chrom1": pairs_df.chrom1.values,
            "start1": region_starts1,
            "end1": region_ends1,
            "chrom2": pairs_df.chrom2.values,
            "start2": region_starts2,
            "end2": region_ends2,
            "strand1": pairs_df.strand1.values,
            "strand2": pairs_df.strand2.values,
            "dist_bin_idx": np.searchsorted(
                dist_bins, np.abs(pairs_df.pos1 - pairs_df.pos2), side="right"
            ),
            "n_pairs": 1,
        },
        copy=False,
    )

    pairs_reduced_df["min_dist"] = np.where(
        pairs_reduced_df["dist_bin_idx"] > 0,
        dist_bins[pairs_reduced_df["dist_bin_idx"] - 1],
        0,
    )

    pairs_reduced_df["max_dist"] = np.where(
        pairs_reduced_df["dist_bin_idx"] < len(dist_bins)-1,
        dist_bins[pairs_reduced_df["dist_bin_idx"]],
        np.iinfo(np.int64).max,
    )

    # importantly, in the future, we may want to extend the function to plot scalings
    # for pairs from different regions!

    pairs_for_scaling_mask = (
        (pairs_reduced_df.chrom1 == pairs_reduced_df.chrom2)
        & (pairs_reduced_df.start1 == pairs_reduced_df.start2)
        & (pairs_reduced_df.end1 == pairs_reduced_df.end2)
        & (pairs_reduced_df.min_dist > 0)
        & (pairs_reduced_df.max_dist < np.iinfo(np.int64).max)
    )

    pairs_for_scaling_df = pairs_reduced_df.loc[pairs_for_scaling_mask]

    pairs_for_scaling_counts = pairs_for_scaling_df.groupby(
        by=[
            "chrom1",
            "start1",
            "end1",
            "chrom2",
            "start2",
            "end2",
            "strand1",
            "strand2",
            "min_dist",
            "max_dist",
        ]
    ).agg({"n_pairs": "sum"})

    pairs_for_scaling_counts = (
        make_empty_scaling(regions, dist_bins)
        .assign(n_pairs=0)
        .add(pairs_for_scaling_counts, fill_value=0)
    )
    pairs_for_scaling_counts["n_pairs"] = pairs_for_scaling_counts["n_pairs"].astype(
        np.int64
    )

    if ignore_trans:
        pairs_no_scaling_counts = None
    else:
        pairs_no_scaling_df = pairs_reduced_df.loc[~pairs_for_scaling_mask]

        pairs_no_scaling_counts = pairs_no_scaling_df.groupby(
            by=[
                "chrom1",
                "start1",
                "end1",
                "chrom2",
                "start2",
                "end2",
                "strand1",
                "strand2",
            ]
        ).agg({"n_pairs": "sum"})

        pairs_no_scaling_counts = (
            make_empty_cross_region_table(regions)
            .assign(n_pairs=0)
            .add(pairs_no_scaling_counts, fill_value=0)
        )
        pairs_no_scaling_counts["n_pairs"] = pairs_no_scaling_counts["n_pairs"].astype(
            np.int64
        )

    return pairs_for_scaling_counts, pairs_no_scaling_counts


def contact_areas_same_reg(min_dist, max_dist, region_length):

    min_dist = _to_float(min_dist)
    max_dist = _to_float(max_dist)
    scaffold_length = _to_float(region_length)
    outer_areas = np.maximum(region_length - min_dist, 0) ** 2
    inner_areas = np.maximum(region_length - max_dist, 0) ** 2
    return 0.5 * (outer_areas - inner_areas)


def _contact_areas_diff_reg(
    min_dist, max_dist, region_start1, region_end1, region_start2, region_end2
):

    return (
        contact_areas_same_reg(min_dist, max_dist, np.abs(region_end2 - region_start1))
        + contact_areas_same_reg(
            min_dist, max_dist, np.abs(region_end1 - region_start2)
        )
        - contact_areas_same_reg(
            min_dist, max_dist, np.abs(region_start1 - region_start2)
        )
        - contact_areas_same_reg(min_dist, max_dist, np.abs(region_end1 - region_end2))
    )


def _contact_areas_trans(min_dist, max_dist, region_length1, region_length2):

    return (
        contact_areas_same_reg(min_dist, max_dist, region_length1 + region_length2)
        - contact_areas_same_reg(min_dist, max_dist, region_length1)
        - contact_areas_same_reg(min_dist, max_dist, region_length2)
    )


def compute_scaling(
    pairs,
    regions=None,
    chromsizes=None,
    dist_range=(int(1e1), int(1e9)),
    n_dist_bins=8 * 8,
    chunksize=int(1e7),
    ignore_trans=False,
    filter_f=None,
    nproc_in=1,
    cmd_in=None,
):
    """
    Main function for computing scaling.

    Parameters
    ----------
    pairs: pd.DataFrame, stream of fiel paht with pairs.
    regions: bioframe viewframe, anything that can serve as input to bioframe.from_any, or None
    chromsizes: additional dataframe with chromosome sizes, if different from regions
    dist_range: (int, int) tuple with distance ranges that will be split into windows
    n_dist_bins: number of logarithmic bins
    chunksize: size of chunks for calculations
    ignore_trans: bool, ignore trans or not
    filter_f: filter function that can be applied to each chunk
    nproc_in
    cmd_in

    Returns
    -------

    """

    dist_bins = geomspace(dist_range[0], dist_range[1], n_dist_bins)

    if isinstance(pairs, pd.DataFrame):
        pairs_df = pairs

    elif isinstance(pairs, str) or hasattr(pairs, "buffer") or hasattr(pairs, "peek"):
        from . import fileio, headerops

        pairs_stream = (
            fileio.auto_open(
                pairs,
                mode="r",
                nproc=nproc_in,
                command=cmd_in,
            )
            if isinstance(pairs, str)
            else pairs
        )

        header, pairs_body = headerops.get_header(pairs_stream)

        cols = headerops.extract_column_names(header)
        if chromsizes is None:
            chromsizes = headerops.extract_chromsizes(header)
        pairs_df = pd.read_csv(
            pairs_body,
            header=None,
            names=cols,
            chunksize=chunksize,
            sep="\t",
            dtype={"chrom1": str, "chrom2": str},
        )
    else:
        raise ValueError(
            "pairs must be either a path to a pairs file or a pd.DataFrame"
        )

    sc, trans_counts = None, None
    for pairs_chunk in [pairs_df] if isinstance(pairs_df, pd.DataFrame) else pairs_df:
        if filter_f:
            pairs_chunk = filter_f(pairs_chunk)
        sc_chunk, trans_counts_chunk = bins_pairs_by_distance(
            pairs_chunk,
            dist_bins,
            regions=regions,
            chromsizes=chromsizes,
            ignore_trans=ignore_trans,
        )

        sc = sc_chunk if sc is None else sc.add(sc_chunk, fill_value=0)
        trans_counts = (
            trans_counts_chunk
            if trans_counts is None
            else trans_counts.add(trans_counts_chunk, fill_value=0)
        )

    #         if not (isinstance(regions, pd.DataFrame) and
    #                  (set(regions.columns) == set(['chrom', 'start','end']))):
    #             raise ValueError('regions must be provided as a dict or chrom-indexed Series of chromsizes or as a bedframe.')

    sc.reset_index(inplace=True)
    sc["n_bp2"] = contact_areas_same_reg(
        sc["min_dist"], sc["max_dist"], sc["end1"] - sc["start1"]
    )

    if not ignore_trans:
        trans_counts.reset_index(inplace=True)
        trans_counts["np_bp2"] = (trans_counts["end1"] - trans_counts["start1"]) * (
            trans_counts["end2"] - trans_counts["start2"]
        )

    return sc, trans_counts


def norm_scaling_factor(bins, cfreqs, anchor=1.0, binwindow=(0, 3)):
    i = np.searchsorted(bins, anchor)
    return cfreqs[i + binwindow[0] : i + binwindow[1]].mean()


def norm_scaling(bins, cfreqs, anchor=1.0, binwindow=(0, 3)):
    return cfreqs / norm_scaling_factor(bins, cfreqs, anchor, binwindow)


def unity_norm_scaling(bins, cfreqs, norm_range=(1e4, 1e9)):
    bin_lens = np.diff(bins)
    bin_mids = np.sqrt(bins[1:] * bins[:-1])

    if norm_range is None:
        norm_cfreqs = cfreqs / np.sum(1.0 * (bin_lens * cfreqs)[np.isfinite(cfreqs)])
    else:
        norm_cfreqs = cfreqs / np.sum(
            1.0
            * (bin_lens * cfreqs)[
                np.isfinite(cfreqs)
                & (bin_mids > norm_range[0])
                & (bin_mids < norm_range[1])
            ]
        )

    return norm_cfreqs



# -*- coding: utf-8 -*-
import os
import sys
import subprocess
import pytest
import tempfile

testdir = os.path.dirname(os.path.realpath(__file__))
mock_pairsam_path = os.path.join(testdir, "data", "mock.pairsam")

tmpdir = tempfile.TemporaryDirectory()
tmpdir_name = tmpdir.name
pairs_path = os.path.join(tmpdir_name, "out.pairs")
sam_path = os.path.join(tmpdir_name, "out.sam")


@pytest.fixture
def setup_split():
    try:
        subprocess.check_output(
            [
                "python",
                "-m",
                "pairtools",
                "split",
                mock_pairsam_path,
                "--output-pairs",
                pairs_path,
                "--output-sam",
                sam_path,
            ],
        )
    except subprocess.CalledProcessError as e:
        print(e.output)
        print(sys.exc_info())
        raise e


def test_split(setup_split):

    pairsam_lines = [l.strip() for l in open(mock_pairsam_path, "r") if l.strip()]
    pairs_lines = [l.strip() for l in open(pairs_path, "r") if l.strip()]
    sam_lines = [l.strip() for l in open(sam_path, "r") if l.strip()]

    # check that all entries survived splitting:
    n_pairsam = len([l for l in pairsam_lines if not l.startswith("#")])
    n_pairs = len([l for l in pairs_lines if not l.startswith("#")])
    n_sam = len([l for l in sam_lines if not l.startswith("@")]) // 2

    assert n_pairsam == n_pairs
    assert n_pairsam == n_sam

    # check that the header survived splitting:
    pairsam_header = [
        l.strip()
        for l in open(mock_pairsam_path, "r")
        if l.strip() and l.startswith("#")
    ]
    pairs_header = [
        l.strip() for l in open(pairs_path, "r") if l.strip() and l.startswith("#")
    ]
    sam_header = [
        l.strip() for l in open(sam_path, "r") if l.strip() and l.startswith("@")
    ]
    assert all(
        any(l in l2 for l2 in pairsam_header)
        for l in sam_header
        if not l.startswith("@PG")
    )
    assert all(
        l in pairsam_header
        for l in pairs_header
        if (not (l.startswith("#columns") or l.startswith("#samheader")))
    )
    columns_pairsam = [l for l in pairsam_header if l.startswith("#columns")][
        0
    ].split()[1:]
    columns_pairs = [l for l in pairs_header if l.startswith("#columns")][0].split()[1:]
    assert (
        ("sam1" in columns_pairsam)
        and ("sam2" in columns_pairsam)
        and ("sam1" not in columns_pairs)
        and ("sam2" not in columns_pairs)
    )
    assert [c for c in columns_pairsam if c != "sam1" and c != "sam2"] == columns_pairs

    tmpdir.cleanup()
